#!/usr/bin/env python
"""
Read Structure input file, Make Feff input file, and run Feff
"""

import os
import sys
from pathlib import Path
import numpy as np
np.seterr(all='ignore')

import wx
import wx.lib.scrolledpanel as scrolled
import wx.lib.agw.flatnotebook as fnb

from xraydb.chemparser import chemparse
from xraydb import atomic_number

import larch
from larch.xafs import feff8l, feff6l
from larch.utils import unixpath, mkdir, read_textfile
from larch.utils.strutils import fix_filename, unique_name, strict_ascii
from larch.site_config import user_larchdir

from larch.wxlib import (LarchFrame, FloatSpin, EditableListBox,
                         FloatCtrl, SetTip, get_icon, SimpleText, pack,
                         Button, Popup, HLine, FileSave, FileOpen, Choice,
                         Check, MenuItem, CEN, LEFT, FRAMESTYLE,
                         Font, FONTSIZE, flatnotebook, LarchUpdaterDialog,
                         PeriodicTablePanel, FeffResultsPanel, LarchWxApp,
                         ExceptionPopup, set_color)


from larch.xrd import structure2feff

LEFT = wx.ALIGN_LEFT
CEN |=  wx.ALL
FNB_STYLE = fnb.FNB_NO_X_BUTTON|fnb.FNB_SMART_TABS
FNB_STYLE |= fnb.FNB_NO_NAV_BUTTONS|fnb.FNB_NODRAG

MAINSIZE = (850, 750)

class Structure2FeffFrame(wx.Frame):
    _about = """Larch structure browser for generating and running Feff.

    Ryuichi Shimogawa <ryuichi.shimogawa@stonybrook.edu>
    """
    def __init__(self, parent=None, _larch=None, path_importer=None, filename=None, **kws):
        wx.Frame.__init__(self, parent, -1, size=MAINSIZE, style=FRAMESTYLE)

        title = "Larch FEFF Input Generator and FEFF Runner"

        self.larch = _larch
        if _larch is None:
            self.larch = larch.Interpreter()
        self.larch.eval("# started Structure browser\n")
        self.larch.eval("if not hasattr('_sys', '_feffruns'): _sys._feffruns = {}")
        self.path_importer = path_importer
        self.subframes = {}
        self.current_structure = None
        self.SetTitle(title)
        self.SetSize(MAINSIZE)
        self.SetFont(Font(FONTSIZE))
        self.createMainPanel()
        self.createMenus()

        self.feff_folder = Path(user_larchdir, 'feff').as_posix()
        mkdir(self.feff_folder)

        self.runs_list = []
        for fname in os.listdir(self.feff_folder):
            full = Path(self.feff_folder, fname).absolute()
            if full.is_dir():
                self.runs_list.append(full.name)

        self.statusbar = self.CreateStatusBar(2, style=wx.STB_DEFAULT_STYLE)
        self.statusbar.SetStatusWidths([-3, -1])
        statusbar_fields = [" ", ""]
        for i in range(len(statusbar_fields)):
            self.statusbar.SetStatusText(statusbar_fields[i], i)
        self.Show()

    def createMainPanel(self):
        display0 = wx.Display(0)
        client_area = display0.ClientArea
        xmin, ymin, xmax, ymax = client_area
        xpos = int((xmax-xmin)*0.07) + xmin
        ypos = int((ymax-ymin)*0.09) + ymin
        self.SetPosition((xpos, ypos))

        # main panel with scrolled panel
        scrolledpanel = scrolled.ScrolledPanel(self)
        panel = wx.Panel(scrolledpanel)
        sizer = wx.GridBagSizer(2,2)

        wids = self.wids = {}

        folderlab = SimpleText(panel, ' Feff Folder: ')
        wids['run_folder'] = wx.TextCtrl(panel, value='calc1', size=(250, -1))

        wids['run_feff'] = Button(panel, ' Run Feff ',
                                  action=self.onRunFeff)
        wids['run_feff'].Disable()
        wids['without_h'] = Check(panel, default=True, label='Remove H atoms',
                                  action=self.onGetFeff)


        wids['central_atom'] = Choice(panel, choices=['<empty>'], size=(80, -1),
                                      action=self.onCentralAtom)
        wids['edge']         = Choice(panel, choices=['K', 'L3', 'L2', 'L1',
                                                      'M5', 'M4'],
                                      size=(80, -1),
                                      action=self.onGetFeff)

        wids['feffvers']      = Choice(panel, choices=['6', '8'], default=1,
                                       size=(80, -1),
                                      action=self.onGetFeff)
        wids['site']         = Choice(panel, choices=['1', '2', '3', '4'],
                                      size=(80, -1),
                                      action=self.onGetFeff)
        wids['cluster_size'] = FloatSpin(panel, value=7.0, digits=2,
                                         increment=0.1, max_val=10,
                                         action=self.onGetFeff)
        wids['central_atom'].Disable()
        wids['edge'].Disable()
        wids['cluster_size'].Disable()
        catomlab = SimpleText(panel, ' Absorbing Atom: ')
        sitelab  = SimpleText(panel, ' Crystal Site: ')
        edgelab  = SimpleText(panel, ' Edge: ')
        csizelab = SimpleText(panel, ' Cluster Size (\u212B): ')
        fverslab = SimpleText(panel, ' Feff Version:')

        ir = 1

        sizer.Add(catomlab,             (ir, 0), (1, 1), LEFT, 3)
        sizer.Add(wids['central_atom'], (ir, 1), (1, 1), LEFT, 3)
        sizer.Add(sitelab,              (ir, 2), (1, 1), LEFT, 3)
        sizer.Add(wids['site'],         (ir, 3), (1, 1), LEFT, 3)
        sizer.Add(edgelab,              (ir, 4), (1, 1), LEFT, 3)
        sizer.Add(wids['edge'],         (ir, 5), (1, 1), LEFT, 3)

        ir += 1
        sizer.Add(csizelab,             (ir, 0), (1, 1), LEFT, 3)
        sizer.Add(wids['cluster_size'], (ir, 1), (1, 1), LEFT, 3)
        sizer.Add(fverslab,             (ir, 2), (1, 1), LEFT, 3)
        sizer.Add(wids['feffvers'],     (ir, 3), (1, 1), LEFT, 3)
        sizer.Add(wids['without_h'],    (ir, 4), (1, 2), LEFT, 3)

        ir += 1
        sizer.Add(folderlab,             (ir, 0), (1, 1), LEFT, 3)
        sizer.Add(wids['run_folder'],    (ir, 1), (1, 4), LEFT, 3)
        sizer.Add(wids['run_feff'],      (ir, 5), (1, 1), LEFT, 3)

        pack(panel, sizer)

        self.nb = flatnotebook(scrolledpanel, {}, on_change=self.onNBChanged)

        self.feffresults = FeffResultsPanel(scrolledpanel,
                                            path_importer=self.path_importer,
                                            _larch=self.larch)

        structure_panel = wx.Panel(scrolledpanel)
        wids['structure_text'] = wx.TextCtrl(structure_panel, value='<STRUCTURE TEXT>',
                                       style=wx.TE_MULTILINE|wx.TE_READONLY,
                                       size=(300, 350))
        wids['structure_text'].SetFont(Font(FONTSIZE+1))
        structure_sizer = wx.BoxSizer(wx.VERTICAL)
        structure_sizer.Add(wids['structure_text'], 1, LEFT|wx.GROW, 1)
        pack(structure_panel, structure_sizer)

        feff_panel = wx.Panel(scrolledpanel)
        wids['feff_text'] = wx.TextCtrl(feff_panel,
                                       value='<Feff Input Text>',
                                       style=wx.TE_MULTILINE,
                                       size=(300, 350))
        wids['feff_text'].CanCopy()

        feff_panel.onPanelExposed = self.onGetFeff
        wids['feff_text'].SetFont(Font(FONTSIZE+1))
        feff_sizer = wx.BoxSizer(wx.VERTICAL)
        feff_sizer.Add(wids['feff_text'], 1, LEFT|wx.GROW, 1)
        pack(feff_panel, feff_sizer)

        feffout_panel = wx.Panel(scrolledpanel)
        wids['feffout_text'] = wx.TextCtrl(feffout_panel,
                                           value='<Feff Output>',
                                           style=wx.TE_MULTILINE,
                                           size=(300, 350))
        wids['feffout_text'].CanCopy()
        wids['feffout_text'].SetFont(Font(FONTSIZE+1))
        feffout_sizer = wx.BoxSizer(wx.VERTICAL)
        feffout_sizer.Add(wids['feffout_text'], 1, LEFT|wx.GROW, 1)
        pack(feffout_panel, feffout_sizer)

        self.nbpages = []
        for label, page in (('Structure Text',  structure_panel),
                            ('Feff Input Text', feff_panel),
                            ('Feff Output Text', feffout_panel),
                            ('Feff Results',    self.feffresults),
                            ):
            self.nb.AddPage(page, label, True)
            self.nbpages.append((label, page))
        self.nb.SetSelection(0)

        r_sizer = wx.BoxSizer(wx.VERTICAL)
        r_sizer.Add(panel, 0, LEFT|wx.GROW|wx.ALL)
        r_sizer.Add(self.nb, 1, LEFT|wx.GROW, 2)
        pack(scrolledpanel, r_sizer)
        scrolledpanel.SetupScrolling()

    def get_nbpage(self, name):
        "get nb page by name"
        name = name.lower()
        for i, dat in enumerate(self.nbpages):
            label, page = dat
            if name in label.lower():
                return i, page
        return (0, self.npbages[0][1])

    def onCentralAtom(self, event=None):
        structure  = self.current_structure
        if structure is None:
            return
        catom = event.GetString()
        try:
            sites = structure2feff.structure_sites(structure['structure_text'], absorber=catom, fmt=structure['fmt'])
            sites = ['%d' % (i+1) for i in range(len(sites))]
            self.wids['site'].Clear()
            self.wids['site'].AppendItems(sites)
            self.wids['site'].Select(0)
        except:
            self.write_message(f"could not get sites for central atom '{catom}'")
            title = f"Could not get sites for central atom '{catom}'"
            message = []
            ExceptionPopup(self, title, message)

        edge_val = 'K' if atomic_number(catom) < 60 else 'L3'
        self.wids['edge'].SetStringSelection(edge_val)
        self.onGetFeff()

    def onGetFeff(self, event=None):
        structure  = self.current_structure
        if structure is None:
            return
        edge  = self.wids['edge'].GetStringSelection()
        version8 = '8' == self.wids['feffvers'].GetStringSelection()
        catom = self.wids['central_atom'].GetStringSelection()
        asite = int(self.wids['site'].GetStringSelection())
        csize = self.wids['cluster_size'].GetValue()
        with_h = not self.wids['without_h'].IsChecked()
        folder = f'{catom:s}{asite:d}_{edge:s}'
        folder = unique_name(fix_filename(folder), self.runs_list)

        fefftext = structure2feff.structure2feffinp(structure['structure_text'], catom, edge=edge,
                                                    cluster_size=csize,
                                                    absorber_site=asite,
                                                    version8=version8,
                                                    with_h=with_h,
                                                    fmt=structure['fmt'])

        self.wids['run_folder'].SetValue(folder)
        self.wids['feff_text'].SetValue(fefftext)
        self.wids['run_feff'].Enable()
        i, p = self.get_nbpage('Feff Input')
        self.nb.SetSelection(i)

    def onRunFeff(self, event=None):
        fefftext = self.wids['feff_text'].GetValue()
        if len(fefftext) < 100 or 'ATOMS' not in fefftext:
            return

        structure_text = self.wids['structure_text'].GetValue()
        structure  = self.current_structure
        structure_fname = None

        if structure is not None:
            structure_fname = structure['fname']

        version8 = '8' == self.wids['feffvers'].GetStringSelection()

        fname = self.wids['run_folder'].GetValue()
        fname = unique_name(fix_filename(fname), self.runs_list)
        self.runs_list.append(fname)
        folder = Path(self.feff_folder, fname).absolute()
        mkdir(folder)

        ix, p = self.get_nbpage('Feff Output')
        self.nb.SetSelection(ix)

        self.folder = folder.as_posix()
        out = self.wids['feffout_text']
        out.Clear()
        out.SetInsertionPoint(0)
        out.WriteText(f'########\n###\n# Run Feff in folder: {folder:s}\n')
        out.SetInsertionPoint(out.GetLastPosition())
        out.WriteText('###\n########\n')
        out.SetInsertionPoint(out.GetLastPosition())

        fname = Path(folder, 'feff.inp').absolute()
        with open(fname, 'w', encoding=sys.getdefaultencoding()) as fh:
            fh.write(strict_ascii(fefftext))

        if structure_fname is not None:
            cname = Path(folder, structure_fname).absolute()
            with open(cname, 'w', encoding=sys.getdefaultencoding()) as fh:
                fh.write(strict_ascii(structure_text))

        wx.CallAfter(self.run_feff, self.folder, version8=version8)

    def run_feff(self, folder, version8=True):
        folder = Path(folder).absolute()
        dname = folder.name
        prog, cmd = feff8l, 'feff8l'
        if not version8:
            prog, cmd = feff6l, 'feff6l'
        command = f"{cmd:s}(folder='{folder}')"
        self.larch.eval(f"## running Feff as:\n#  {command:s}\n##\n")

        prog(folder=folder.as_posix(), message_writer=self.feff_output)
        self.larch.eval("## gathering results:\n")
        self.larch.eval(f"_sys._feffruns['{dname}'] = get_feff_pathinfo('{folder}')")
        this_feffrun = self.larch.symtable._sys._feffruns[f'{dname}']
        self.feffresults.set_feffresult(this_feffrun)
        ix, p = self.get_nbpage('Feff Results')
        self.nb.SetSelection(ix)

        # clean up unused, intermediate Feff files
        for fname in os.listdir(folder):
            if (fname.endswith('.json') or fname.endswith('.pad') or
                fname.endswith('.bin') or fname.startswith('log') or
                fname in ('chi.dat', 'xmu.dat', 'misc.dat')):
                os.unlink(Path(folder, fname).absolute())

    def feff_output(self, text):
        out = self.wids['feffout_text']
        ix, p = self.get_nbpage('Feff Output')
        self.nb.SetSelection(ix)
        pos0 = out.GetLastPosition()
        if not text.endswith('\n'):
            text = '%s\n' % text
        out.WriteText(text)
        out.SetInsertionPoint(out.GetLastPosition())
        out.Update()
        out.Refresh()

    def onExportFeff(self, event=None):
        if self.current_structure is None:
            return
        fefftext = self.wids['feff_text'].GetValue()
        if len(fefftext) < 20:
            return
        cc = self.current_structure
        fname = f'{cc["fname"]}_feff.inp'
        wildcard = 'Feff Inut files (*.inp)|*.inp|All files (*.*)|*.*'
        path = FileSave(self, message='Save Feff File',
                        wildcard=wildcard,
                        default_file=fname)
        if path is not None:
            with open(path, 'w', encoding=sys.getdefaultencoding()) as fh:
                fh.write(fefftext)
            self.write_message("Wrote Feff file %s" % path, 0)

    def onExportStructure(self, event=None):
        if self.current_structure is None:
            return

        cc = self.current_structure
        fname = cc["fname"]
        wildcard = f'Sturcture files (*.{cc["fmt"]})|*.{cc["fmt"]}|All files (*.*)|*.*'
        path = FileSave(self, message='Save Structure File',
                        wildcard=wildcard,
                        default_file=fname)

        if path is not None:
            with open(path, 'w', encoding=sys.getdefaultencoding()) as fh:
                fh.write(cc['structure_text'])
            self.write_message("Wrote structure file %s" % path, 0)

    def onImportStructure(self, event=None):
        wildcard = 'Strucuture files (*.cif/*.postcar/*.contcar/*.chgcar/*locpot/*.cssr)|*.cif;*.postcar;*.contcar;*.chgcar;*locpot;*.cssr|Molecule files (*.xyz/*.gjf/*.g03/*.g09/*.com/*.inp)|*.xyz;*.gjf;*.g03;*.g09;*.com;*.inp|All other files readable with Openbabel (*.*)|*.*'
        path = FileOpen(self, message='Open Structure File',
                        wildcard=wildcard, default_file='My.cif')

        if path is not None:
            fmt = path.split('.')[-1]
            fname = Path(path).name
            with open(path, 'r', encoding=sys.getdefaultencoding()) as f:
                structure_text = f.read()

            self.current_structure = structure2feff.parse_structure(structure_text=structure_text, fmt=fmt, fname=fname)

        self.wids['structure_text'].SetValue(self.current_structure['structure_text'])

        # use pytmatgen to get formula
        elems =  chemparse(self.current_structure['formula'].replace(' ', ''))

        self.wids['central_atom'].Enable()
        self.wids['edge'].Enable()
        self.wids['cluster_size'].Enable()

        self.wids['central_atom'].Clear()
        self.wids['central_atom'].AppendItems(list(elems.keys()))
        self.wids['central_atom'].Select(0)



        el0 = list(elems.keys())[0]
        edge_val = 'K' if atomic_number(el0) < 60 else 'L3'
        self.wids['edge'].SetStringSelection(edge_val)

        # sites
        sites = structure2feff.structure_sites(self.current_structure['structure_text'], fmt=self.current_structure["fmt"], absorber=el0)
        try:
            sites = ['%d' % (i+1) for i in range(len(sites))]
        except:
            title = "Could not make sense of atomic sites"
            message = [f"Elements: {list(elems.keys())}",
                       f"Sites: {sites}"]
            ExceptionPopup(self, title, message)


        self.wids['site'].Clear()
        self.wids['site'].AppendItems(sites)
        self.wids['site'].Select(0)
        i, p = self.get_nbpage('Structure Text')
        self.nb.SetSelection(i)

    def onImportFeff(self, event=None):
        wildcard = 'Feff input files (*.inp)|*.inp|All files (*.*)|*.*'
        path = FileOpen(self, message='Open Feff Input File',
                        wildcard=wildcard, default_file='feff.inp')
        if path is not None:
            fefftext = None
            fname = Path(path).name.replace('.inp', '_run')
            fname = unique_name(fix_filename(fname), self.runs_list)
            fefftext = read_textfile(path)
            if fefftext is not None:
                self.wids['feff_text'].SetValue(fefftext)
                self.wids['run_folder'].SetValue(fname)
                self.wids['run_feff'].Enable()
                i, p = self.get_nbpage('Feff Input')
                self.nb.SetSelection(i)

    def onFeffFolder(self, eventa=None):
        "prompt for Feff Folder"
        dlg = wx.DirDialog(self, 'Select Main Folder for Feff Calculations',
                           style=wx.DD_DEFAULT_STYLE|wx.DD_CHANGE_DIR)

        dlg.SetPath(self.feff_folder)
        if  dlg.ShowModal() == wx.ID_CANCEL:
            return None
        self.feff_folder = Path(dlg.GetPath()).absolute().as_posix()
        mkdir(self.feff_folder)

    def onNBChanged(self, event=None):
        callback = getattr(self.nb.GetCurrentPage(), 'onPanelExposed', None)
        if callable(callback):
            callback()

    def onSelAll(self, event=None):
        self.controller.filelist.select_all()

    def onSelNone(self, event=None):
        self.controller.filelist.select_none()

    def write_message(self, msg, panel=0):
        """write a message to the Status Bar"""
        self.statusbar.SetStatusText(msg, panel)

    def createMenus(self):
        self.menubar = wx.MenuBar()
        fmenu = wx.Menu()
        group_menu = wx.Menu()
        data_menu = wx.Menu()
        ppeak_menu = wx.Menu()
        m = {}

        MenuItem(self, fmenu, "&Open Structure File\tCtrl+O",
                 "Open Structure File",  self.onImportStructure)

        MenuItem(self, fmenu, "&Save Structure File\tCtrl+S",
                 "Save Structure File",  self.onExportStructure)

        MenuItem(self, fmenu, "Open Feff Input File",
                 "Open Feff input File",  self.onImportFeff)

        MenuItem(self, fmenu, "Save &Feff Inp File\tCtrl+F",
                 "Save Feff6 File",  self.onExportFeff)

        fmenu.AppendSeparator()
        MenuItem(self, fmenu, "Select Main Feff Folder",
                 "Select Main Folder for running Feff",
                 self.onFeffFolder)
        fmenu.AppendSeparator()
        MenuItem(self, fmenu, "Quit",  "Exit", self.onClose)

        self.menubar.Append(fmenu, "&File")

        self.SetMenuBar(self.menubar)
        self.Bind(wx.EVT_CLOSE,  self.onClose)

    def onClose(self, event=None):
        self.Destroy()


class Structure2FeffViewer(LarchWxApp):
    def __init__(self, filename=None, version_info=None,  **kws):
        self.filename = filename
        LarchWxApp.__init__(self, version_info=version_info, **kws)

    def createApp(self):
        frame = Structure2FeffFrame(filename=self.filename,
                         version_info=self.version_info)
        self.SetTopWindow(frame)
        return True

def structure_viewer(**kws):
    Structure2FeffViewer(**kws)

if __name__ == '__main__':
    Structure2FeffViewer().MainLoop()



# Useful physical constants
# most of these are put into common X-ray units (Angstroms, ev)

import scipy.constants as consts
from numpy import pi

I = 0.0 + 1.0j

RAD2DEG = 180.0/pi
DEG2RAD = pi/180.0
PI = pi
TAU = 2*pi

# cross-section unit
BARN     = 1.e-24   # cm^2

# atoms/mol =  6.0221413e23  atoms/mol
AVOGADRO = consts.Avogadro

# ATOMIC MASS in grams
AMU = consts.atomic_mass * 1000.0

# electron rest mass in eV
E_MASS = consts.electron_mass * consts.c**2 / consts.e

# Planck's Constant
#   h*c    ~= 12398.42 eV*Ang
#   hbar*c ~=  1973.27 eV*Ang
PLANCK_HC    = 1.e10 * consts.Planck * consts.c / consts.e
PLANCK_HBARC = PLANCK_HC / TAU

# Rydberg constant in eV (~13.6 eV)
RYDBERG = consts.Rydberg * consts.Planck * consts.c/ consts.e

# classical electron radius in cm and Ang
R_ELECTRON_CM  = 100.0 * consts.physical_constants['classical electron radius'][0]
R_ELECTRON_ANG = 1.e8 * R_ELECTRON_CM


# a few standard lattice constants
STD_LATTICE_CONSTANTS = {'Si': 5.4310205, 'C': 3.567095, 'Ge': 5.64613}


# will be able to import these from xraydb when v 4.5.1 is required
ATOM_SYMS = ['H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na', 'Mg',
           'Al', 'Si', 'P', 'S', 'Cl', 'Ar', 'K', 'Ca', 'Sc', 'Ti', 'V', 'Cr',
           'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn', 'Ga', 'Ge', 'As', 'Se', 'Br',
           'Kr', 'Rb', 'Sr', 'Y', 'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd',
           'Ag', 'Cd', 'In', 'Sn', 'Sb', 'Te', 'I', 'Xe', 'Cs', 'Ba', 'La',
           'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 'Er',
           'Tm', 'Yb', 'Lu', 'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt', 'Au',
           'Hg', 'Tl', 'Pb', 'Bi', 'Po', 'At', 'Rn', 'Fr', 'Ra', 'Ac', 'Th',
           'Pa', 'U', 'Np', 'Pu', 'Am', 'Cm', 'Bk', 'Cf', 'Es', 'Fm', 'Md',
           'No', 'Lr', 'Rf', 'Db', 'Sg', 'Bh', 'Hs', 'Mt', 'Ds', 'Rg', 'Cn',
           'Nh', 'Fl', 'Mc', 'Lv', 'Ts', 'Og']

ATOM_NAMES = ['hydrogen', 'helium', 'lithium', 'beryllium', 'boron', 'carbon',
            'nitrogen', 'oxygen', 'fluorine', 'neon', 'sodium', 'magnesium',
            'aluminum', 'silicon', 'phosphorus', 'sulfur', 'chlorine', 'argon',
            'potassium', 'calcium', 'scandium', 'titanium', 'vanadium',
            'chromium', 'manganese', 'iron', 'cobalt', 'nickel', 'copper',
            'zinc', 'gallium', 'germanium', 'arsenic', 'selenium', 'bromine',
            'krypton', 'rubidium', 'strontium', 'yttrium', 'zirconium',
            'niobium', 'molybdenum', 'technetium', 'ruthenium', 'rhodium',
            'palladium', 'silver', 'cadmium', 'indium', 'tin', 'antimony',
            'tellurium', 'iodine', 'xenon', 'cesium', 'barium', 'lanthanum',
            'cerium', 'praseodymium', 'neodymium', 'promethium', 'samarium',
            'europium', 'gadolinium', 'terbium', 'dysprosium', 'holmium',
            'erbium', 'thulium', 'ytterbium', 'lutetium', 'hafnium',
            'tantalum', 'tungsten', 'rhenium', 'osmium', 'iridium', 'platinum',
            'gold', 'mercury', 'thallium', 'lead', 'bismuth', 'polonium',
            'astatine', 'radon', 'francium', 'radium', 'actinium', 'thorium',
            'protactinium', 'uranium', 'neptunium', 'plutonium', 'americium',
            'curium', 'berkelium', 'californium', 'einsteinium', 'fermium',
            'mendelevium', 'nobelium', 'lawrencium', 'rutherfordium',
            'dubnium', 'seaborgium', 'bohrium', 'hassium', 'meitnerium',
            'darmstadtium', 'roentgenium', 'copernicium', 'nihonium',
            'flerovium', 'moscovium', 'livermorium', 'tennessine', 'oganesson']



from typing import NamedTuple, List, Optional
from numpy.typing import ArrayLike


class XasScan(NamedTuple):
    name: str
    description: str
    info: str
    start_time: str
    labels: List[str]
    data: ArrayLike


class XasDataSource:
    TYPE = NotImplemented

    def __init__(self, filename: str) -> None:
        self._filename = filename

    def get_source_info(self) -> str:
        raise NotImplementedError

    def get_scan(self, scan_name: str) -> Optional[XasScan]:
        raise NotImplementedError

    def get_scan_names(self) -> List[str]:
        raise NotImplementedError

    def get_sorted_scan_names(self) -> List[str]:
        scan_names = self.get_scan_names()
        return sorted(scan_names, key=lambda s: float(s) if s.isdigit() else s)



#!/usr/bin/env python
"""
Browse CIF Files, maybe run Feff
"""

import os
import sys
import time
import copy
# from threading import Thread
import numpy as np
np.seterr(all='ignore')
from pathlib import Path
from functools import partial
import wx
import wx.lib.scrolledpanel as scrolled
import wx.lib.agw.flatnotebook as fnb
from wx.adv import AboutBox, AboutDialogInfo
from matplotlib.ticker import FuncFormatter

from wxmplot import PlotPanel
from xraydb.chemparser import chemparse
from xraydb import atomic_number

import larch
from larch import Group
from larch.xafs import feff8l, feff6l
from larch.xrd.cif2feff import cif_sites
from larch.utils import read_textfile, mkdir
from larch.utils.paths import unixpath
from larch.utils.strutils import fix_filename, unique_name, strict_ascii
from larch.site_config import user_larchdir

from larch.wxlib import (LarchFrame, FloatSpin, EditableListBox,
                         FloatCtrl, SetTip, get_icon, SimpleText, pack,
                         Button, Popup, HLine, FileSave, FileOpen, Choice,
                         Check, MenuItem, CEN, LEFT, FRAMESTYLE,
                         Font, FONTSIZE, flatnotebook, LarchUpdaterDialog,
                         PeriodicTablePanel, FeffResultsPanel, LarchWxApp,
                         ExceptionPopup, set_color)

from larch.xrd import CifStructure, get_amcsd, find_cifs, get_cif, parse_cif_file

LEFT = wx.ALIGN_LEFT
CEN |=  wx.ALL
FNB_STYLE = fnb.FNB_NO_X_BUTTON|fnb.FNB_SMART_TABS
FNB_STYLE |= fnb.FNB_NO_NAV_BUTTONS|fnb.FNB_NODRAG

MAINSIZE = (1150, 650)

class CIFFrame(wx.Frame):
    _about = """Larch Crystallographic Information File Browser
    Data from American Mineralogist Crystal Structure Database

    Matt Newville <newville @ cars.uchicago.edu>
    """

    def __init__(self, parent=None, _larch=None, with_feff=False,
                 with_fdmnes=False, usecif_callback=None, path_importer=None,
                 filename=None, **kws):

        wx.Frame.__init__(self, parent, -1, size=MAINSIZE, style=FRAMESTYLE)

        title = "Larch American Mineralogist CIF Browser"
        self.with_feff = with_feff
        self.with_fdmnes = with_fdmnes
        self.usecif_callback = usecif_callback
        self.larch = _larch
        if _larch is None:
            self.larch = larch.Interpreter()
        self.larch.eval("# started CIF browser\n")

        self.path_importer = path_importer
        self.cifdb = get_amcsd()
        self.all_minerals = self.cifdb.all_minerals()
        self.subframes = {}
        self.has_xrd1d = False
        self.xrd1d_thread = None
        self.current_cif = None
        self.SetTitle(title)
        self.SetSize(MAINSIZE)
        self.SetFont(Font(FONTSIZE))

        self.createMainPanel()
        self.createMenus()

        if with_feff:
            self.larch.eval("if not hasattr('_sys', '_feffruns'): _sys._feffruns = {}")
            self.feff_folder = unixpath(Path(user_larchdir, 'feff'))
            mkdir(self.feff_folder)
            self.feffruns_list = []
            for fname in os.listdir(self.feff_folder):
                full = Path(self.feff_folder, fname).absolute()
                if full.is_dir():
                    self.feffruns_list.append(fname)

        self.statusbar = self.CreateStatusBar(2, style=wx.STB_DEFAULT_STYLE)
        self.statusbar.SetStatusWidths([-3, -1])
        statusbar_fields = [" ", ""]
        for i in range(len(statusbar_fields)):
            self.statusbar.SetStatusText(statusbar_fields[i], i)
        self.Show()

    def createMainPanel(self):
        display0 = wx.Display(0)
        client_area = display0.ClientArea
        xmin, ymin, xmax, ymax = client_area
        xpos = int((xmax-xmin)*0.07) + xmin
        ypos = int((ymax-ymin)*0.09) + ymin
        self.SetPosition((xpos, ypos))

        splitter  = wx.SplitterWindow(self, style=wx.SP_LIVE_UPDATE)
        splitter.SetMinimumPaneSize(250)

        leftpanel = wx.Panel(splitter, size=(375, -1))
        self.ciflist = EditableListBox(leftpanel, self.onShowCIF, size=(375, -1))
        set_color(self.ciflist, 'list_fg', bg='list_bg')
        self.cif_selections = {}

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.ciflist, 1, LEFT|wx.GROW|wx.ALL, 1)
        pack(leftpanel, sizer)

        # right hand side
        rightpanel = scrolled.ScrolledPanel(splitter)
        panel = wx.Panel(rightpanel, size=(725, -1))

        self.ciflist.SetMinSize((375, 250))
        rightpanel.SetMinSize((400, 250))

        sizer = wx.GridBagSizer(2, 2)

        self.title = SimpleText(panel, 'Search American Mineralogical CIF Database:',
                                size=(700, -1), style=LEFT)
        self.title.SetFont(Font(FONTSIZE+2))
        wids = self.wids = {}

        minlab = SimpleText(panel, ' Mineral Name: ')
        minhint= SimpleText(panel, ' example: hem* ')
        wids['mineral'] = wx.TextCtrl(panel, value='',   size=(250, -1),
                                      style=wx.TE_PROCESS_ENTER)
        wids['mineral'].Bind(wx.EVT_TEXT_ENTER, self.onSearch)

        authlab = SimpleText(panel, ' Author Name: ')
        wids['author'] = wx.TextCtrl(panel, value='',   size=(250, -1),
                                     style=wx.TE_PROCESS_ENTER)
        wids['author'].Bind(wx.EVT_TEXT_ENTER, self.onSearch)

        journlab = SimpleText(panel, ' Journal Name: ')
        wids['journal'] = wx.TextCtrl(panel, value='',   size=(250, -1),
                                      style=wx.TE_PROCESS_ENTER)
        wids['journal'].Bind(wx.EVT_TEXT_ENTER, self.onSearch)

        elemlab = SimpleText(panel, ' Include Elements: ')
        elemhint= SimpleText(panel, ' example: O, Fe, Si ')

        wids['contains_elements'] = wx.TextCtrl(panel, value='', size=(250, -1),
                                                style=wx.TE_PROCESS_ENTER)
        wids['contains_elements'].Bind(wx.EVT_TEXT_ENTER, self.onSearch)

        exelemlab = SimpleText(panel, ' Exclude Elements: ')
        wids['excludes_elements'] = wx.TextCtrl(panel, value='', size=(250, -1),
                                                style=wx.TE_PROCESS_ENTER)
        wids['excludes_elements'].Bind(wx.EVT_TEXT_ENTER, self.onSearch)

        wids['excludes_elements'].Enable()
        wids['strict_contains'] = Check(panel, default=False,
                                       label='Include only the elements listed',
                                       action=self.onStrict)

        wids['full_occupancy'] = Check(panel, default=False,
                                       label='Only Structures with Full Occupancy')

        wids['search']   = Button(panel, 'Search for CIFs',  action=self.onSearch)


        ir = 0
        sizer.Add(self.title,     (0, 0), (1, 6), LEFT, 2)

        ir += 1
        sizer.Add(HLine(panel, size=(650, 2)), (ir, 0), (1, 6), LEFT, 3)

        ir += 1
        sizer.Add(minlab,          (ir, 0), (1, 1), LEFT, 3)
        sizer.Add(wids['mineral'], (ir, 1), (1, 3), LEFT, 3)
        sizer.Add(minhint,         (ir, 4), (1, 2), LEFT, 3)
        ir += 1
        sizer.Add(authlab,        (ir, 0), (1, 1), LEFT, 3)
        sizer.Add(wids['author'], (ir, 1), (1, 3), LEFT, 3)

        ir += 1
        sizer.Add(journlab,        (ir, 0), (1, 1), LEFT, 3)
        sizer.Add(wids['journal'], (ir, 1), (1, 3), LEFT, 3)

        ir += 1
        sizer.Add(elemlab,        (ir, 0), (1, 1), LEFT, 3)
        sizer.Add(wids['contains_elements'], (ir, 1), (1, 3), LEFT, 3)
        sizer.Add(elemhint,         (ir, 4), (1, 3), LEFT, 2)

        ir += 1
        sizer.Add(exelemlab,        (ir, 0), (1, 1), LEFT, 3)
        sizer.Add(wids['excludes_elements'], (ir, 1), (1, 3), LEFT, 3)

        ir += 1
        sizer.Add(wids['search'],          (ir, 0), (1, 1), LEFT, 3)
        sizer.Add(wids['strict_contains'], (ir, 1), (1, 4), LEFT, 3)

        ir += 1
        sizer.Add(wids['full_occupancy'], (ir, 1), (1, 4), LEFT, 3)

        #
        if self.with_feff:
            wids['feff_runfolder'] = wx.TextCtrl(panel, value='calc1', size=(250, -1))
            wids['feff_runbutton'] = Button(panel, ' Run Feff ', action=self.onRunFeff)
            wids['feff_runbutton'].Disable()
            wids['feff_without_h'] = Check(panel, default=True, label='Remove H atoms',
                                           action=self.onGetFeff)


            wids['feff_central_atom'] = Choice(panel, choices=['<empty>'], size=(80, -1),
                                          action=self.onFeffCentralAtom)
            wids['feff_edge']         = Choice(panel, choices=['K', 'L3', 'L2', 'L1',
                                                          'M5', 'M4'],
                                          size=(80, -1),
                                          action=self.onGetFeff)

            wids['feffvers']      = Choice(panel, choices=['6', '8'], default=1,
                                           size=(80, -1),
                                          action=self.onGetFeff)
            wids['feff_site']         = Choice(panel, choices=['1', '2', '3', '4'],
                                          size=(80, -1),
                                          action=self.onGetFeff)
            wids['feff_cluster_size'] = FloatSpin(panel, value=7.0, digits=2,
                                             increment=0.1, max_val=10,
                                             action=self.onGetFeff)
            wids['feff_central_atom'].Disable()
            wids['feff_edge'].Disable()
            wids['feff_cluster_size'].Disable()

            ir += 1
            sizer.Add(HLine(panel, size=(650, 2)), (ir, 0), (1, 6), LEFT, 3)

            ir += 1

            sizer.Add(SimpleText(panel, ' Absorbing Atom: '),  (ir, 0), (1, 1), LEFT, 3)
            sizer.Add(wids['feff_central_atom'], (ir, 1), (1, 1), LEFT, 3)
            sizer.Add(SimpleText(panel, ' Crystal Site: '), (ir, 2), (1, 1), LEFT, 3)
            sizer.Add(wids['feff_site'],     (ir, 3), (1, 1), LEFT, 3)
            sizer.Add(SimpleText(panel, ' Edge: '),  (ir, 4), (1, 1), LEFT, 3)
            sizer.Add(wids['feff_edge'],     (ir, 5), (1, 1), LEFT, 3)

            ir += 1
            sizer.Add(SimpleText(panel, ' Cluster Size (\u212B): '),  (ir, 0), (1, 1), LEFT, 3)
            sizer.Add(wids['feff_cluster_size'], (ir, 1), (1, 1), LEFT, 3)
            sizer.Add(SimpleText(panel, ' Feff Version:'),     (ir, 2), (1, 1), LEFT, 3)
            sizer.Add(wids['feffvers'],     (ir, 3), (1, 1), LEFT, 3)
            sizer.Add(wids['feff_without_h'],    (ir, 4), (1, 2), LEFT, 3)

            ir += 1
            sizer.Add(SimpleText(panel, ' Feff Folder: '),      (ir, 0), (1, 1), LEFT, 3)
            sizer.Add(wids['feff_runfolder'],    (ir, 1), (1, 4), LEFT, 3)
            sizer.Add(wids['feff_runbutton'],      (ir, 5), (1, 1), LEFT, 3)

        if self.usecif_callback is not None:
            wids['cif_use_button'] = Button(panel, ' Use This CIF', action=self.onUseCIF)
            wids['cif_use_button'].Disable()

            ir += 1
            sizer.Add(wids['cif_use_button'],    (ir, 5), (1, 1), LEFT, 3)


        ir += 1
        sizer.Add(HLine(panel, size=(650, 2)), (ir, 0), (1, 6), LEFT, 3)

        pack(panel, sizer)

        self.nb = flatnotebook(rightpanel, {}, on_change=self.onNBChanged)


        def _swallow_plot_messages(s, panel=0):
            pass

        self.plotpanel = PlotPanel(rightpanel, messenger=_swallow_plot_messages)
        try:
            plotopts = self.larch.symtable._sys.wx.plotopts
            self.plotpanel.conf.set_theme(plotopts['theme'])
            self.plotpanel.conf.enable_grid(plotopts['show_grid'])
        except:
            pass

        self.plotpanel.SetMinSize((250, 250))
        self.plotpanel.SetMaxSize((675, 400))
        self.plotpanel.onPanelExposed = self.showXRD1D

        cif_panel = wx.Panel(rightpanel)
        wids['cif_text'] = wx.TextCtrl(cif_panel, value='<CIF TEXT>',
                                       style=wx.TE_MULTILINE|wx.TE_READONLY,
                                       size=(700, 450))
        wids['cif_text'].SetFont(Font(FONTSIZE+1))
        cif_sizer = wx.BoxSizer(wx.VERTICAL)
        cif_sizer.Add(wids['cif_text'], 0, LEFT, 1)
        pack(cif_panel, cif_sizer)


        self.nbpages = []
        for label, page in (('CIF Text',  cif_panel),
                            ('1-D XRD Pattern', self.plotpanel),
                            ):
            self.nb.AddPage(page, label, True)
            self.nbpages.append((label, page))

        if self.with_feff:
            self.feffresults = FeffResultsPanel(rightpanel,
                                                path_importer=self.path_importer,
                                                _larch=self.larch)

            feffinp_panel = wx.Panel(rightpanel)
            wids['feff_text'] = wx.TextCtrl(feffinp_panel,
                                           value='<Feff Input Text>',
                                           style=wx.TE_MULTILINE,
                                           size=(700, 450))
            wids['feff_text'].CanCopy()

            feffinp_panel.onPanelExposed = self.onGetFeff
            wids['feff_text'].SetFont(Font(FONTSIZE+1))
            feff_sizer = wx.BoxSizer(wx.VERTICAL)
            feff_sizer.Add(wids['feff_text'], 0, LEFT, 1)
            pack(feffinp_panel, feff_sizer)

            feffout_panel = wx.Panel(rightpanel)
            wids['feffout_text'] = wx.TextCtrl(feffout_panel,
                                               value='<Feff Output>',
                                               style=wx.TE_MULTILINE,
                                               size=(700, 450))
            wids['feffout_text'].CanCopy()
            wids['feffout_text'].SetFont(Font(FONTSIZE+1))
            feffout_sizer = wx.BoxSizer(wx.VERTICAL)
            feffout_sizer.Add(wids['feffout_text'], 0, LEFT, 1)
            pack(feffout_panel, feffout_sizer)

            for label, page in (('Feff Input Text', feffinp_panel),
                                ('Feff Output Text', feffout_panel),
                                ('Feff Results',    self.feffresults),
                                ):
                self.nb.AddPage(page, label, True)
                self.nbpages.append((label, page))
        self.nb.SetSelection(0)

        r_sizer = wx.BoxSizer(wx.VERTICAL)
        r_sizer.Add(panel, 0, LEFT|wx.GROW|wx.ALL)
        r_sizer.Add(self.nb, 1, LEFT|wx.GROW, 2)

        pack(rightpanel, r_sizer)
        rightpanel.SetupScrolling()
        splitter.SplitVertically(leftpanel, rightpanel, 1)


    def get_nbpage(self, name):
        "get nb page by name"
        name = name.lower()
        for i, dat in enumerate(self.nbpages):
            label, page = dat
            if name in label.lower():
                return i, page
        return (0, self.nbpages[0][1])

    def onStrict(self, event=None):
        strict = self.wids['strict_contains'].IsChecked()
        self.wids['excludes_elements'].Enable(not strict)

    def onSearch(self, event=None):
        mineral_name = self.wids['mineral'].GetValue().strip()
        if len(mineral_name) < 1:
            mineral_name = None
        author_name = self.wids['author'].GetValue().strip()
        if len(author_name) < 1:
            author_name = None
        journal_name = self.wids['journal'].GetValue().strip()
        if len(journal_name) < 1:
            journal_name = None
        contains_elements = self.wids['contains_elements'].GetValue().strip()
        if len(contains_elements) < 1:
            contains_elements = None
        else:
            contains_elements = [a.strip().title() for a in contains_elements.split(',')]
        excludes_elements = self.wids['excludes_elements'].GetValue().strip()
        if len(excludes_elements) < 1:
            excludes_elements = None
        else:
            excludes_elements = [a.strip().title() for a in excludes_elements.split(',')]
        strict_contains = self.wids['strict_contains'].IsChecked()
        full_occupancy = self.wids['full_occupancy'].IsChecked()
        all_cifs = find_cifs(mineral_name=mineral_name,
                             journal_name=journal_name,
                             author_name=author_name,
                             contains_elements=contains_elements,
                             excludes_elements=excludes_elements,
                             strict_contains=strict_contains,
                             full_occupancy=full_occupancy)
        if len(all_cifs) == 0:
            all_cifs = find_cifs(mineral_name=mineral_name + '*',
                                 journal_name=journal_name,
                                 author_name=author_name,
                                 contains_elements=contains_elements,
                                 excludes_elements=excludes_elements,
                                 strict_contains=strict_contains,
                                 full_occupancy=full_occupancy)
        self.cif_selections = {}
        self.ciflist.Clear()
        for cif in all_cifs:
            try:
                label = cif.formula.replace(' ', '')
                mineral = cif.get_mineralname()
                year = cif.publication.year
                journal= cif.publication.journalname
                cid = cif.ams_id
                label = f'{label}: {mineral}, {year} {journal} [{cid}]'
            except:
                label = None
            if label is not None:
                if label in self.cif_selections:
                    lorig, n = label, 1
                    while label in self.cif_selections and n < 10:
                        n += 1
                        label = f'{lorig} (v{n})'

                self.cif_selections[label] =  cif.ams_id
                self.ciflist.Append(label)

    def onShowCIF(self, event=None, cif_id=None):
        if cif_id is not None:
            cif = get_cif(cif_id)
            self.cif_label = '%d' % cif_id
        elif event is not None:
            self.cif_label = event.GetString()
            cif = get_cif(self.cif_selections[self.cif_label])
        self.current_cif = cif
        self.has_xrd1d = False
        self.wids['cif_text'].SetValue(cif.ciftext)

        if self.with_feff:
            elems =  chemparse(cif.formula.replace(' ', ''))
            self.wids['feff_central_atom'].Enable()
            self.wids['feff_edge'].Enable()
            self.wids['feff_cluster_size'].Enable()

            self.wids['feff_central_atom'].Clear()
            self.wids['feff_central_atom'].AppendItems(list(elems.keys()))
            self.wids['feff_central_atom'].Select(0)

            el0 = list(elems.keys())[0]
            edge_val = 'K' if atomic_number(el0) < 60 else 'L3'
            self.wids['feff_edge'].SetStringSelection(edge_val)

            sites = cif_sites(cif.ciftext, absorber=el0)
            try:
                sites = ['%d' % (i+1) for i in range(len(sites))]
            except:
                title = "Could not make sense of atomic sites"
                message = [f"Elements: {list(elems.keys())}",
                           f"Sites: {sites}"]
                ExceptionPopup(self, title, message)

            self.wids['feff_site'].Clear()
            self.wids['feff_site'].AppendItems(sites)
            self.wids['feff_site'].Select(0)

        if self.usecif_callback is not None:
            self.wids['cif_use_button'].Enable()

        i, p = self.get_nbpage('CIF Text')
        self.nb.SetSelection(i)

    def onUseCIF(self, event=None):
        if self.usecif_callback is not None:
            self.usecif_callback(cif=self.current_cif)


    def onFeffCentralAtom(self, event=None):
        cif  = self.current_cif
        if cif is None:
            return
        catom = event.GetString()
        try:
            sites = cif_sites(cif.ciftext, absorber=catom)
            sites = ['%d' % (i+1) for i in range(len(sites))]
            self.wids['feff_site'].Clear()
            self.wids['feff_site'].AppendItems(sites)
            self.wids['feff_site'].Select(0)
        except:
            self.write_message(f"could not get sites for central atom '{catom}'")
            title = f"Could not get sites for central atom '{catom}'"
            message = []
            ExceptionPopup(self, title, message)

        edge_val = 'K' if atomic_number(catom) < 60 else 'L3'
        self.wids['feff_edge'].SetStringSelection(edge_val)
        self.onGetFeff()

    def onGetFeff(self, event=None):
        cif  = self.current_cif
        if cif is None or not self.with_feff:
            return

        edge  = self.wids['feff_edge'].GetStringSelection()
        version8 = '8' == self.wids['feffvers'].GetStringSelection()
        catom = self.wids['feff_central_atom'].GetStringSelection()
        asite = int(self.wids['feff_site'].GetStringSelection())
        csize = self.wids['feff_cluster_size'].GetValue()
        with_h = not self.wids['feff_without_h'].IsChecked()
        mineral = cif.get_mineralname()
        folder = f'{catom:s}{asite:d}_{edge:s}_{mineral}_cif{cif.ams_id:d}'
        folder = unique_name(fix_filename(folder), self.feffruns_list)

        fefftext = cif.get_feffinp(catom, edge=edge, cluster_size=csize,
                                   absorber_site=asite, version8=version8,
                                   with_h=with_h)

        self.wids['feff_runfolder'].SetValue(folder)
        self.wids['feff_text'].SetValue(fefftext)
        self.wids['feff_runbutton'].Enable()
        i, p = self.get_nbpage('Feff Input')
        self.nb.SetSelection(i)

    def onRunFeff(self, event=None):
        fefftext = self.wids['feff_text'].GetValue()
        if len(fefftext) < 100 or 'ATOMS' not in fefftext or not self.with_feff:
            return

        ciftext = self.wids['cif_text'].GetValue()
        cif  = self.current_cif
        cif_fname = None
        if cif is not None and len(ciftext) > 100:
            mineral = cif.get_mineralname()
            cif_fname = f'{mineral}_cif{cif.ams_id:d}.cif'

        # cc = self.current_cif
        # edge  = self.wids['feff_edge'].GetStringSelection()
        # catom = self.wids['feff_central_atom'].GetStringSelection()
        # asite = int(self.wids['feff_site'].GetStringSelection())
        # mineral = cc.get_mineralname()
        # folder = f'{catom:s}{asite:d}_{edge:s}_{mineral}_cif{cc.ams_id:d}'
        # folder = unixpath(Path(self.feff_folder, folder))
        version8 = '8' == self.wids['feffvers'].GetStringSelection()

        fname = self.wids['feff_runfolder'].GetValue()
        fname = unique_name(fix_filename(fname), self.feffruns_list)
        self.feffruns_list.append(fname)
        self.folder = folder = unixpath(Path(self.feff_folder, fname))
        mkdir(self.folder)
        ix, p = self.get_nbpage('Feff Output')
        self.nb.SetSelection(ix)

        out = self.wids['feffout_text']
        out.Clear()
        out.SetInsertionPoint(0)
        out.WriteText(f'########\n###\n# Run Feff in folder: {folder}\n')
        out.SetInsertionPoint(out.GetLastPosition())
        out.WriteText('###\n########\n')
        out.SetInsertionPoint(out.GetLastPosition())

        fname = unixpath(Path(folder, 'feff.inp').absolute())
        with open(fname, 'w', encoding=sys.getdefaultencoding()) as fh:
            fh.write(strict_ascii(fefftext))

        if cif_fname is not None:
            cname = unixpath(Path(folder, fix_filename(cif_fname)))
            with open(cname, 'w', encoding=sys.getdefaultencoding()) as fh:
                fh.write(strict_ascii(ciftext))
        wx.CallAfter(self.run_feff, folder, version8=version8)

    def run_feff(self, folder=None, version8=True):
        # print("RUN FEFF ", folder)
        dname = Path(folder).name
        prog, cmd = feff8l, 'feff8l'
        if not version8:
            prog, cmd = feff6l, 'feff6l'
        command = f"{cmd:s}(folder='{folder:s}')"
        self.larch.eval(f"## running Feff as:\n#  {command:s}\n##\n")

        prog(folder=folder, message_writer=self.feff_output)
        self.larch.eval("## gathering results:\n")
        self.larch.eval(f"_sys._feffruns['{dname:s}'] = get_feff_pathinfo('{folder:s}')")
        this_feffrun = self.larch.symtable._sys._feffruns[f'{dname:s}']
        self.feffresults.set_feffresult(this_feffrun)
        ix, p = self.get_nbpage('Feff Results')
        self.nb.SetSelection(ix)

        # clean up unused, intermediate Feff files
        for fname in os.listdir(folder):
            if (fname.endswith('.json') or fname.endswith('.pad') or
                fname.endswith('.bin') or fname.startswith('log') or
                fname in ('chi.dat', 'xmu.dat', 'misc.dat')):
                os.unlink(unixpath(Path(folder, fname).absolute()))

    def feff_output(self, text):
        out = self.wids['feffout_text']
        ix, p = self.get_nbpage('Feff Output')
        self.nb.SetSelection(ix)
        pos0 = out.GetLastPosition()
        if not text.endswith('\n'):
            text = '%s\n' % text
        out.WriteText(text)
        out.SetInsertionPoint(out.GetLastPosition())
        out.Update()
        out.Refresh()

    def onExportFeff(self, event=None):
        if self.current_cif is None:
            return
        fefftext = self.wids['feff_text'].GetValue()
        if len(fefftext) < 20:
            return
        cc = self.current_cif
        minname = cc.get_mineralname()
        fname = f'{minname}_cif{cc.ams_id:d}_feff.inp'
        wildcard = 'Feff Inut files (*.inp)|*.inp|All files (*.*)|*.*'
        path = FileSave(self, message='Save Feff File',
                        wildcard=wildcard,
                        default_file=fname)
        path = unixpath(path)
        if path is not None:
            with open(path, 'w', encoding=sys.getdefaultencoding()) as fh:
                fh.write(fefftext)
            self.write_message("Wrote Feff file %s" % path, 0)


    def onExportCIF(self, event=None):
        if self.current_cif is None:
            return
        cc = self.current_cif
        minname = cc.get_mineralname()
        fname = f'{minname}_cif{cc.ams_id:d}.cif'
        wildcard = 'CIF files (*.cif)|*.cif|All files (*.*)|*.*'
        path = FileSave(self, message='Save CIF File',
                        wildcard=wildcard,
                        default_file=fname)
        path = unixpath(path)
        if path is not None:
            with open(path, 'w', encoding=sys.getdefaultencoding()) as fh:
                fh.write(cc.ciftext)
            self.write_message("Wrote CIF file %s" % path, 0)

    def onImportCIF(self, event=None):
        wildcard = 'CIF files (*.cif)|*.cif|All files (*.*)|*.*'
        path = FileOpen(self, message='Open CIF File',
                        wildcard=wildcard, default_file='My.cif')
        path = unixpath(path)
        if path is not None:
            try:
                cif_data = parse_cif_file(path)
            except:
                title = f"Cannot parse CIF file '{path}'"
                message = [f"Error reading CIF File: {path}"]
                ExceptionPopup(self, title, message)
                return

            try:
                cif_id = self.cifdb.add_ciffile(path)
            except:
                title = f"Cannot add CIF from '{path}' to CIF database"
                message = [f"Error adding CIF File to database: {path}"]
                ExceptionPopup(self, title, message)
                return

            try:
                self.onShowCIF(cif_id=cif_id)
            except:
                title = f"Cannot show CIF from '{path}'"
                message = [f"Error displaying CIF File: {path}"]
                ExceptionPopup(self, title, message)

    def onImportFeff(self, event=None):
        if not self.with_feff:
            return
        wildcard = 'Feff input files (*.inp)|*.inp|All files (*.*)|*.*'
        path = FileOpen(self, message='Open Feff Input File',
                        wildcard=wildcard, default_file='feff.inp')
        path = unixpath(path)
        if path is not None:
            fefftext = None
            fname = Path(path).name
            fname = fname.replace('.inp', '_run')
            fname = unique_name(fix_filename(fname), self.feffruns_list)
            fefftext = read_textfile(path)
            if fefftext is not None:
                self.wids['feff_text'].SetValue(fefftext)
                self.wids['feff_runfolder'].SetValue(fname)
                self.wids['feff_runbutton'].Enable()
                i, p = self.get_nbpage('Feff Input')
                self.nb.SetSelection(i)

    def onFeffFolder(self, eventa=None):
        "prompt for Feff Folder"
        dlg = wx.DirDialog(self, 'Select Main Folder for Feff Calculations',
                           style=wx.DD_DEFAULT_STYLE|wx.DD_CHANGE_DIR)

        dlg.SetPath(self.feff_folder)
        if  dlg.ShowModal() == wx.ID_CANCEL:
            return None
        self.feff_folder = Path(dlg.GetPath()).absolute().as_posix()
        mkdir(self.feff_folder)

    def onNBChanged(self, event=None):
        callback = getattr(self.nb.GetCurrentPage(), 'onPanelExposed', None)
        if callable(callback):
            callback()

    def showXRD1D(self, event=None):
        if self.has_xrd1d or self.current_cif is None:
            return

        def display_xrd1d():
            t0 = time.time()
            sfact = self.current_cif.get_structure_factors()
            try:
                self.cifdb.set_hkls(self.current_cif.ams_id, sfact.hkls)
            except:
                pass

            max_ = sfact.intensity.max()
            mask = np.where(sfact.intensity>max_/10.0)[0]
            qval = sfact.q[mask]
            ival = sfact.intensity[mask]
            ival = ival/(1.0*ival.max())

            def qd_formatter(q, pos):
                qval = float(q)
                dval = '\n[%.2f]' % (2*np.pi/max(qval, 1.e-6))
                return r"%.2f%s" % (qval, dval)

            qd_label = r'$Q\rm\,(\AA^{-1}) \,\> [d \rm\,(\AA)]$'
            title = self.cif_label + '\n' + '(cif %d)' % (self.current_cif.ams_id)
            ppan = self.plotpanel
            ppan.plot(qval, ival, linewidth=0, marker='o', markersize=2,
                      xlabel=qd_label, ylabel='Relative Intensity',
                      title=title, titlefontsize=8, delay_draw=True)

            ppan.axes.bar(qval, ival, 0.1, color='blue')
            ppan.axes.xaxis.set_major_formatter(FuncFormatter(qd_formatter))
            ppan.canvas.draw()
            self.has_xrd1d = True

        display_xrd1d()
#         self.xrd1d_thread = Thread(target=display_xrd1d)
#         self.xrd1d_thread.start()
#         time.sleep(0.25)
#         self.xrd1d_thread.join()


    def onSelAll(self, event=None):
        self.controller.filelist.select_all()

    def onSelNone(self, event=None):
        self.controller.filelist.select_none()

    def write_message(self, msg, panel=0):
        """write a message to the Status Bar"""
        self.statusbar.SetStatusText(msg, panel)

    def createMenus(self):
        # ppnl = self.plotpanel
        self.menubar = wx.MenuBar()
        fmenu = wx.Menu()
        group_menu = wx.Menu()
        data_menu = wx.Menu()
        ppeak_menu = wx.Menu()
        m = {}

        MenuItem(self, fmenu, "&Open CIF File\tCtrl+O",
                 "Open CIF File",  self.onImportCIF)

        MenuItem(self, fmenu, "&Save CIF File\tCtrl+S",
                 "Save CIF File",  self.onExportCIF)

        MenuItem(self, fmenu, "Open Feff Input File",
                 "Open Feff input File",  self.onImportFeff)

        MenuItem(self, fmenu, "Save &Feff Inp File\tCtrl+F",
                 "Save Feff6 File",  self.onExportFeff)

        fmenu.AppendSeparator()
        MenuItem(self, fmenu, "Select Main Feff Folder",
                 "Select Main Folder for running Feff",
                 self.onFeffFolder)
        fmenu.AppendSeparator()
        MenuItem(self, fmenu, "Quit",  "Exit", self.onClose)

        self.menubar.Append(fmenu, "&File")

        self.SetMenuBar(self.menubar)
        self.Bind(wx.EVT_CLOSE,  self.onClose)

    def onClose(self, event=None):
        self.Destroy()


class CIFViewer(LarchWxApp):
    def __init__(self, filename=None, version_info=None,  **kws):
        self.filename = filename
        LarchWxApp.__init__(self, version_info=version_info, **kws)

    def createApp(self):
        frame = CIFFrame(filename=self.filename,
                         version_info=self.version_info)
        self.SetTopWindow(frame)
        return True

def cif_viewer(**kws):
    CIFViewer(**kws)

if __name__ == '__main__':
    CIFViewer().MainLoop()



#!/usr/bin/env python
#

import sys
import os
import time
from functools import partial
from pathlib import Path
import wx
import wx.lib.mixins.inspection

import numpy
import scipy
import larch
from pyshortcuts import platform

from wxutils import (MenuItem, Font, Button, Choice, panel_pack)

from .gui_utils import LarchWxApp
from .readlinetextctrl import ReadlineTextCtrl
from .larchfilling import Filling
from .columnframe import ColumnDataFileFrame
from .athena_importer import AthenaImporter
from . import inputhook

from larch.io import (read_ascii, read_xdi, read_gsexdi,
                      gsescan_group,
                      is_athena_project, AthenaProject)
from larch.version import make_banner, version_data
from larch.utils import get_cwd, fix_varname

FILE_WILDCARDS = "Data Files(*.0*,*.dat,*.xdi)|*.0*;*.dat;*.xdi|All files (*.*)|*.*"

ICON_FILE = 'larch.ico'
BACKGROUND_COLOUR = '#FCFCFA'
FOREGROUND_COLOUR = '#050520'

FONTSIZE_FW = 14
if platform == 'win':
    FONTSIZE_FW = 12
elif platform == 'darwin':
    FONTSIZE_FW = 14

def makeColorPanel(parent, color):
    p = wx.Panel(parent, -1)
    p.SetBackgroundColour(color)
    return p

def wx_inspect():
    wx.GetApp().ShowInspectionTool()

def get_font(size):
    return wx.Font(size, wx.FONTFAMILY_SWISS, wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_BOLD)

class LarchWxShell(object):
    ps1 = 'Larch>'
    ps2 = ' ... >'
    def __init__(self, wxparent=None,   writer=None, _larch=None,
                 prompt=None, historyfile=None, output=None, input=None):
        self._larch = _larch
        self.textstyle = None
        self.parent = wxparent
        self.prompt = prompt
        self.input  = input
        self.output = output

        if _larch is None:
            self._larch  = larch.Interpreter(historyfile=historyfile,
                                             writer=self)
            self._larch.run_init_scripts()
        self.writer  = self._larch.writer
        self.symtable = self._larch.symtable

        self.objtree = wxparent.objtree

        self.set_textstyle(mode='text')
        self._larch("_sys.display.colors['text2'] = {'color': 'blue'}",
                    add_history=False)

        self.symtable.set_symbol('_builtin.force_wxupdate', False)
        self.symtable.set_symbol('_sys.wx.inputhook',   inputhook)
        self.symtable.set_symbol('_sys.wx.ping',   inputhook.ping)
        self.symtable.set_symbol('_sys.wx.force_wxupdate', False)
        self.symtable.set_symbol('_sys.wx.wx_inspect', wx_inspect)
        self.symtable.set_symbol('_sys.wx.wxapp', wx.GetApp())
        self.symtable.set_symbol('_sys.wx.parent', wx.GetApp().GetTopWindow())
        self.symtable.set_symbol('_sys.last_eval_time', 0.0)
        self.fontsize = FONTSIZE_FW

        if self.output is not None:
            style = self.output.GetDefaultStyle()
            bgcol = style.GetBackgroundColour()
            sfont = style.GetFont()
            sfont.Family = wx.TELETYPE
            sfont.Weight = wx.BOLD
            sfont.PointSize = self.fontsize
            style.SetFont(sfont)
            self.output.SetDefaultStyle(style)
            self.textstyle = wx.TextAttr('black', bgcol, sfont)
        self.SetPrompt(True)

    def onUpdate(self, event=None):
        symtable = self.symtable
        if symtable.get_symbol('_builtin.force_wxupdate', create=True):
            app = wx.GetApp()
            evtloop = wx.EventLoop()
            while evtloop.Pending():
                evtloop.Dispatch()
            app.ProcessIdle()
        symtable.set_symbol('_builtin.force_wxupdate', False)


    def SetPrompt(self, complete):
        if self.prompt is None:
            return
        sprompt, scolor = self.ps1, '#000075'
        if not complete:
            sprompt, scolor = self.ps2, '#E00075'
        self.prompt.SetLabel(sprompt)
        self.prompt.SetForegroundColour(scolor)
        self.prompt.Refresh()

    def set_textstyle(self, mode='text'):
        if self.output is None:
            return

        display_colors = self.symtable._sys.display.colors

        textattrs = display_colors.get(mode, {'color':'black'})
        color = textattrs['color']

        style = self.output.GetDefaultStyle()
        bgcol = BACKGROUND_COLOUR
        sfont = self.output.GetFont()
        style.SetFont(sfont)
        self.output.SetDefaultStyle(style)
        self.textstyle = wx.TextAttr(color, bgcol, sfont)

    def write_sys(self, text):
        sys.stdout.write(text)
        sys.stdout.flush()

    def write(self, text, **kws):
        if text is None:
            return
        if self.textstyle is None:
            self.set_textstyle()

        if self.output is None or self.textstyle is None:
            self.write_sys(text)
        else:
            self.output.SetInsertionPointEnd()
            pos0 = self.output.GetLastPosition()
            self.output.WriteText(text)
            pos1 = self.output.GetLastPosition()
            self.output.SetStyle(pos0, pos1, self.textstyle)
            self.output.SetInsertionPoint(pos1)
        self.output.Refresh()
        wx.CallAfter(self.input.SetFocus)

    def flush(self, *args):
        self.output.Refresh()
        self.needs_flush = False

    def clear_input(self):
        self._larch.input.clear()
        self.SetPrompt(True)

    def onFlushTimer(self, event=None):
        if self.needs_flush:
            self.flush()

    def eval(self, text, add_history=True, **kws):
        if text is None:
            return
        if text.startswith('!'):
            return os.system(text[1:])

        elif text.startswith('help(') and text.endswith(')'):
            topic = text[5:-1]
            parent = self.symtable.get_parentpath(topic)
            self.objtree.ShowNode("%s.%s" % (parent, topic))
            return
        else:
            if add_history:
                self.parent.AddToHistory(text)
                self.write("%s\n" % text)
            ret = self._larch.eval(text, add_history=add_history)
            if self._larch.error:
                self._larch.input.clear()
                self._larch.writer.set_textstyle('error')
                self._larch.show_errors()
                self._larch.writer.set_textstyle('text')
            elif ret is not None:
                self._larch.writer.write("%s\n" % repr(ret))
            try:
                self.objtree.onRefresh()
            except ValueError:
                pass
            self.symtable._sys.last_eval_time = time.time()
            self.SetPrompt(self._larch.input.complete)
            return ret

class LarchPanel(wx.Panel):
    """Larch Input/Output Panel + Data Viewer as a wx.Panel,
    suitable for embedding into apps
    """
    def __init__(self,  parent=None, _larch=None, font=None,
                 historyfile='history_larchgui.lar', **kwds):
        self.parent = parent
        if not historyfile.startswith(larch.site_config.user_larchdir):
            historyfile = Path(larch.site_config.user_larchdir,
                                       historyfile).as_posix()

        wx.Panel.__init__(self, parent, -1, size=(750, 725))

        self.splitter = splitter = wx.SplitterWindow(self, style=wx.SP_LIVE_UPDATE)
        splitter.SetMinimumPaneSize(150)

        self.objtree = Filling(splitter,  rootLabel='_main',
                               fgcol=FOREGROUND_COLOUR, bgcol=BACKGROUND_COLOUR)

        self.output = wx.TextCtrl(splitter, -1,  '',
                                  style=wx.TE_MULTILINE|wx.TE_RICH|wx.TE_READONLY)

        self.output.SetBackgroundColour(BACKGROUND_COLOUR)
        self.output.SetForegroundColour(FOREGROUND_COLOUR)
        if font is None:
            font = get_font(self.fontsize)

        self.output.SetFont(font)
        self.objtree.tree.SetFont(font)
        self.objtree.text.SetFont(font)

        self.output.CanCopy()
        self.output.SetInsertionPointEnd()
        splitter.SplitHorizontally(self.objtree, self.output, 0)

        ipanel = wx.Panel(self)

        self.prompt = wx.StaticText(ipanel, label='Larch>', size=(75,-1),
                                    style=wx.ALIGN_CENTER|wx.ALIGN_RIGHT)

        self.input = wx.TextCtrl(ipanel, value='', size=(525,-1),
                                 style=wx.TE_LEFT|wx.TE_PROCESS_ENTER)

        self.input.Bind(wx.EVT_TEXT_ENTER, self.onText)
        if sys.platform == 'darwin':
            self.input.Bind(wx.EVT_KEY_UP,  self.onChar)
        else:
            self.input.Bind(wx.EVT_CHAR,  self.onChar)

        self.hist_buff = []
        self.hist_mark = 0

        isizer = wx.BoxSizer(wx.HORIZONTAL)
        isizer.Add(self.prompt,  0, wx.BOTTOM|wx.CENTER)
        isizer.Add(self.input,   1, wx.ALIGN_LEFT|wx.EXPAND)

        ipanel.SetSizer(isizer)
        isizer.Fit(ipanel)

        opts = dict(flag=wx.ALL|wx.EXPAND, border=2)
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(splitter,  1, **opts)
        sizer.Add(ipanel, 0, **opts)

        self.SetSizer(sizer)
        self.larchshell = LarchWxShell(wxparent=self,
                                       _larch = _larch,
                                       historyfile=historyfile,
                                       prompt = self.prompt,
                                       output = self.output,
                                       input  = self.input)

        self.objtree.SetRootObject(self.larchshell.symtable)
        # root = self.objtree.tree.GetRootItem()


    def write_banner(self):
        self.larchshell.set_textstyle('text2')
        self.larchshell.write(make_banner(show_libraries=['numpy', 'scipy', 'matplotlib', 'h5py',
                                                          'lmfit', 'xraydb', 'wx','wxmplot']))
        self.larchshell.write("\n  \n")
        self.larchshell.set_textstyle('text')

    def update(self):
        self.objtree.onRefresh()

    def onText(self, event=None):
        text =  event.GetString()
        self.input.Clear()
        if text.lower() in ('quit', 'exit', 'quit()', 'exit()'):
            if self.parent.exit_on_close:
                self.parent.onExit()
        else:
            wx.CallAfter(self.larchshell.eval, text)

    def onChar(self, event=None):
        key = event.GetKeyCode()

        entry  = self.input.GetValue().strip()
        pos = self.input.GetSelection()
        ctrl = event.ControlDown()

        if key == wx.WXK_RETURN and len(entry) > 0:
            pass
        if key in (wx.WXK_UP, wx.WXK_DOWN):
            if key == wx.WXK_UP:
                self.hist_mark = max(0, self.hist_mark-1)
            else:
                self.hist_mark += 1
            try:
                wx.CallAfter(self.set_input_text, self.hist_buff[self.hist_mark])
            except IndexError:
                wx.CallAfter(self.set_input_text, '')
        event.Skip()

    def set_input_text(self, text):
        self.input.SetValue(text)
        self.input.SetFocus()
        self.input.SetInsertionPointEnd()


    def AddToHistory(self, text=''):
        for tline in text.split('\n'):
            if len(tline.strip()) > 0:
                self.hist_buff.append(tline)
                self.hist_mark = len(self.hist_buff)


class LarchFrame(wx.Frame):
    def __init__(self, parent=None, _larch=None, is_standalone=True,
                 historyfile='history_larchgui.lar', with_inspection=False,
                 exit_on_close=False, with_raise=True, **kwds):

        self.is_standalone = is_standalone
        self.with_inspection = with_inspection
        self.exit_on_close = exit_on_close
        self.parent = parent
        self.historyfile = historyfile
        self.subframes = {}
        self.last_array_sel = {}
        self.fontsize = FONTSIZE_FW

        wx.Frame.__init__(self, parent, -1, size=(800, 725),
                          style= wx.DEFAULT_FRAME_STYLE)
        self.SetTitle('LarchGUI')

        self.font = get_font(self.fontsize)
        self.SetFont(self.font)
        sbar = self.CreateStatusBar(2, wx.CAPTION)

        self.SetStatusWidths([-2,-1])
        self.SetStatusText("Larch initializing...", 0)

        self.mainpanel = LarchPanel(parent=self, _larch=_larch,
                                    historyfile=historyfile,
                                    font=self.font)

        self.larchshell = self.mainpanel.larchshell
        self._larch = self.larchshell._larch

        sizer = wx.BoxSizer(wx.VERTICAL)

        sizer.Add(self.mainpanel, 1, wx.ALL|wx.EXPAND)

        self.SetSizer(sizer)

        self.Bind(wx.EVT_SHOW, self.onShow)
        self.BuildMenus()
        self.onSelectFont(fsize=self.fontsize)
        # larchdir = larch.site_config.larchdir

        fico = Path(larch.site_config.icondir, ICON_FILE).absolute()
        if fico.exists():
            self.SetIcon(wx.Icon(fico.as_posix(), wx.BITMAP_TYPE_ICO))
        self.mainpanel.write_banner()
        if with_raise:
            self.Raise()

    def Raise(self):
        self.SetStatusText("Ready", 0)
        self.Refresh()
        wx.Frame.Raise(self)

    def BuildMenus(self):
        menuBar = wx.MenuBar()

        fmenu = wx.Menu()
        if self.is_standalone:
            MenuItem(self, fmenu, "&Read Data File\tCtrl+O",
                     "Read Data File", self.onReadData)
        MenuItem(self, fmenu, "&Read and Run Larch Script\tCtrl+R",
                 "Read and Execute a Larch Script", self.onRunScript)
        MenuItem(self, fmenu, "&Save Session History\tCtrl+S",
                 "Save Session History to File", self.onSaveHistory)
        MenuItem(self, fmenu, 'Change Working Directory\tCtrl+W',
                 'Change Directory', self.onChangeDir)
        MenuItem(self, fmenu, 'Clear Input\tCtrl+D',
                 'Clear Input', self.onClearInput)

        if self.with_inspection:
            MenuItem(self, fmenu, 'Show wxPython Inspector\tCtrl+I',
                     'Debug wxPython App', self.onWxInspect)
        fmenu.AppendSeparator()

        if self.parent is None and self.exit_on_close:
            self.Bind(wx.EVT_CLOSE,  self.onExit)
            MenuItem(self, fmenu, 'E&xit', 'End program',
                     self.onExit)
        else:
            self.Bind(wx.EVT_CLOSE,  self.onClose)
            MenuItem(self, fmenu, 'Close Display',
                     'Close display', self.onClose)

        menuBar.Append(fmenu, '&File')

        _sys = self.larchshell.symtable._sys
        if self.is_standalone and hasattr(_sys, 'gui_apps'):
            appmenu = wx.Menu()
            x_apps = _sys.gui_apps.keys()
            for appname in sorted(x_apps):
                label, creator = _sys.gui_apps[appname]

                MenuItem(self, appmenu, label, label,
                         partial(self.show_subframe,
                                 name=appname, creator=creator))
            menuBar.Append(appmenu, 'Applications')

        fsmenu = wx.Menu()
        self.fontsizes = {}
        for fsize in (10, 11, 12, 13, 14, 15, 16, 18, 20, 22, 24):
            m = MenuItem(self, fsmenu,  "%d" % fsize, "%d" % fsize,
                         self.onSelectFont, kind=wx.ITEM_RADIO)
            self.fontsizes[m.GetId()] = fsize
            if fsize == self.fontsize:
                m.Check()

        menuBar.Append(fsmenu, 'Font Size')

        hmenu = wx.Menu()
        MenuItem(self, hmenu, '&About',
                 'Information about this program',  self.onAbout)
        MenuItem(self, hmenu, '&Versions',
                 'Show versions of Larch and libraries',  self.onVersions)
        menuBar.Append(hmenu, '&Help')
        self.SetMenuBar(menuBar)

    def onSelectFont(self, event=None, fsize=None):
        if fsize is None:
            fsize = self.fontsizes.get(event.GetId(), self.fontsize)
        self.fontsize = fsize

        def set_fontsize(obj, fsize):
            fn = obj.GetFont()
            f1, f2 = fn.PixelSize
            fn.SetPixelSize(wx.Size(int((f1*fsize/f2)), fsize))
            obj.SetFont(fn)

        self.PointSize = fsize
        set_fontsize(self, fsize)
        set_fontsize(self.mainpanel.output,  fsize)
        set_fontsize(self.mainpanel.objtree.tree, fsize)
        set_fontsize(self.mainpanel.objtree.text, fsize)
        self.mainpanel.objtree.text.fontsize = fsize



    def onWxInspect(self, event=None):
        wx.GetApp().ShowInspectionTool()

    def onXRFviewer(self, event=None):
        self.larchshell.eval("xrf_plot()")

    def onClearInput(self, event=None):
        self.larchshell.clear_input()

    def onClearInput(self, event=None):
        self.larchshell.clear_input()

    def show_subframe(self, event=None, name=None, creator=None, **opts):
        if name is None or creator is None:
            return
        shown = False
        if name in self.subframes:
            try:
                self.subframes[name].Raise()
                shown = True
            except:
                del self.subframes[name]
        if not shown:
            self.subframes[name] = creator(parent=self,
                                           _larch=self.larchshell._larch,
                                           **opts)
            self.subframes[name].Show()

    def onReadData(self, event=None):
        wildcard = 'Data file (*.dat)|*.dat|All files (*.*)|*.*'
        dlg = wx.FileDialog(self, message='Open Data File',
                            defaultDir=get_cwd(),
                            wildcard=FILE_WILDCARDS,
                            style=wx.FD_OPEN|wx.FD_CHANGE_DIR)
        path = None
        if dlg.ShowModal() == wx.ID_OK:
            path = Path(dlg.GetPath()).absolute().as_posix()
        dlg.Destroy()

        if path is None:
            return

        if is_athena_project(path):
            self.show_subframe(name='athena_import', filename=path,
                               creator=AthenaImporter,
                               read_ok_cb=self.onReadAthenaProject_OK)
        else:
            filename = Path(path).fname
            pref = fix_varname((filename + '_'*8)[:8]).replace('.', '_').lower()

            count, maxcount = 1, 9999
            groupname = "%s%3.3i" % (pref, count)
            while hasattr(self.larchshell.symtable, groupname) and count < maxcount:
                count += 1
                groupname = '%s%3.3i' % (pref, count)

            fh = open(path, 'r')
            line1 = fh.readline().lower()
            fh.close()
            reader = read_ascii
            if 'epics stepscan file' in line1:
                reader = read_gsexdi
            elif 'epics scan' in line1:
                reader = gsescan_group
            elif 'xdi' in line1:
                reader = read_xdi

            dgroup = reader(str(path), _larch=self.larchshell._larch)
            dgroup._path = path
            dgroup._filename = filename
            dgroup._groupname = groupname
            self.show_subframe(name='coledit', event=None,
                               creator=ColumnDataFileFrame,
                               filename=path,
                               last_array_sel=self.last_array_sel,
                               read_ok_cb=self.onReadScan_Success)


    def onReadScan_Success(self, script, path, groupname=None, array_sel=None,
                           overwrite=False):
        """ called when column data has been selected and is ready to be used"""
        self.larchshell.eval(script.format(group=groupname, path=path))
        if array_sel is not None:
            self.last_array_sel = array_sel
        self.larchshell.flush()

    def onReadAthenaProject_OK(self, path, namelist):
        """read groups from a list of groups from an athena project file"""
        read_cmd = "_prj = read_athena('{path:s}', do_fft=False, do_bkg=False)"
        self.larchshell.eval(read_cmd.format(path=path))
        dgroup = None
        script = "{group:s} = _prj.{prjgroup:s}"
        for gname in namelist:
            this = getattr(self.larchshell.symtable._prj, gname)
            gid = str(getattr(this, 'athena_id', gname))
            self.larchshell.eval(script.format(group=gid, prjgroup=gname))
        self.larchshell.eval("del _prj")

    def onRunScript(self, event=None):
        wildcard = 'Larch file (*.lar)|*.lar|All files (*.*)|*.*'
        dlg = wx.FileDialog(self, message='Open and Run Larch Script',
                            wildcard=wildcard,
                            style=wx.FD_OPEN|wx.FD_CHANGE_DIR)
        if dlg.ShowModal() == wx.ID_OK:
            fout = Path(dlg.GetPath()).absolute()
            fname = fout.name
            os.chdir(fout.parent)
            text = "run('%s')" % fname
            self.larchshell.write("%s\n" % text)
            wx.CallAfter(self.larchshell.eval, text)
        dlg.Destroy()

    def onSaveHistory(self, event=None):
        wildcard = 'Larch file (*.lar)|*.lar|All files (*.*)|*.*'
        deffile = 'history.lar'
        dlg = wx.FileDialog(self, message='Save Session History File',
                            wildcard=wildcard,
                            defaultFile=deffile,
                            style=wx.FD_SAVE|wx.FD_CHANGE_DIR)
        if dlg.ShowModal() == wx.ID_OK:
            fout = Path(dlg.GetPath()).absolute().as_posix()
            self._larch.input.history.save(fout, session_only=True)
            self.SetStatusText("Wrote %s" % fout, 0)
        dlg.Destroy()

    def onText(self, event=None):
        text =  event.GetString()
        self.larchshell.write("%s\n" % text)
        self.input.Clear()
        if text.lower() in ('quit', 'exit', 'quit()', 'exit()'):
            if self.exit_on_close:
                self.onExit()
        else:
            self.panel.AddToHistory(text)
            wx.CallAfter(self.larchshell.eval, text)

    def onChangeDir(self, event=None):
        dlg = wx.DirDialog(None, 'Choose a Working Directory',
                           defaultPath = get_cwd(),
                           style = wx.DD_DEFAULT_STYLE)

        if dlg.ShowModal() == wx.ID_OK:
            os.chdir(dlg.GetPath())
        dlg.Destroy()
        return get_cwd()

    def onAbout(self, event=None):
        about_msg =  """LarchGui:
        %s""" % (make_banner(withlibraries=True))

        dlg = wx.MessageDialog(self, about_msg,
                               "About LarchGui", wx.OK | wx.ICON_INFORMATION)
        dlg.ShowModal()
        dlg.Destroy()

    def onVersions(self, event=None):
        vdat = version_data(with_libraries=True)
        out = []
        for key, val in vdat.items():
            out.append(f"{key:20s}:  {val}")
        version_message =  '\n'.join(out)
        dlg = wx.Dialog(self, wx.ID_ANY, size=(700, 400),
                        title='Larch Versions')

        font = get_font(self.fontsize)
        dlg.SetFont(font)
        panel = wx.Panel(dlg)
        txt = wx.StaticText(panel, label=version_message)
        s = wx.Sizer
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(txt, 1, wx.LEFT|wx.ALL, 5)
        panel.SetSizer(sizer)
        panel_pack(dlg, panel)
        dlg.Show()

    def onShow(self, event=None):
        if event.Show:
            self.mainpanel.update()

    def onClose(self, event=None):
        try:
            self.Hide()
        except:
            pass

    def onExit(self, event=None, force=False, with_sysexit=True):
        if not self.exit_on_close:
            self.Hide()
            return
        if force:
            ret = wx.ID_YES
        else:
            dlg = wx.MessageDialog(None, 'Really Quit?', 'Question',
                                   wx.YES_NO | wx.NO_DEFAULT | wx.ICON_QUESTION)
            ret = dlg.ShowModal()

        if ret == wx.ID_YES:
            try:
                self._larch.input.history.save()
            except:
                pass
            try:
                try:
                    for a in self.GetChildren():
                        a.Destroy()
                except:
                    pass
                self.Destroy()

            except:
                pass
            if with_sysexit:
                sys.exit()
        else:
            try:
                event.Veto()
            except:
                pass

class LarchApp(LarchWxApp):
    "simple app to wrap LarchFrame"
    def __init__(self, with_inspection=False, **kws):
        self.with_inspection = with_inspection
        LarchWxApp.__init__(self, **kws)

    def createApp(self):
        frame = LarchFrame(exit_on_close=True, with_inspection=self.with_inspection)
        frame.Show()
        self.SetTopWindow(frame)
        return True

if __name__ == '__main__':
    LarchApp().MainLoop()



from larch import Interpreter
from larch_plugins.xafs.autobk import autobk
from larch_plugins.io.xdi import read_xdi

mylarch = Interpreter()
dat = read_xdi('../xafsdata/fe3c_rt.xdi', _larch=mylarch)
dat.mu = dat.mutrans
autobk(dat, rbkg=1.0, kweight=2, _larch=mylarch)



#!/usr/bin/env python
"""
Larch command-line shell
"""
import cmd
import os
import sys
import signal
from .interpreter import Interpreter

from .site_config import history_file
from .version import make_banner
from .larchlib import StdWriter
from .utils import uname

HAS_READLINE = False
try:
    import readline
    HAS_READLINE = True
except ImportError:
    pass

HAS_WXPYTHON = False
try:
    import wx
    HAS_WXPYTHON = True
except ImportError:
    wx = None


class Shell(cmd.Cmd):
    """command shell for Larch"""
    def __init__(self,  completekey='tab', debug=False, quiet=False,
                 stdin=None, stdout=None, banner_msg=None,
                 maxhist=25000, with_wx=False):

        self.debug  = debug
        cmd.Cmd.__init__(self,completekey='tab')
        if stdin is not None:
            sys.stdin = stdin
        if stdout is not None:
            sys.stdout = stdout
        self.stdin = sys.stdin
        self.stdout = sys.stdout

        if HAS_READLINE:
            try:
                readline.read_history_file(history_file)
            except IOError:
                print(f'could not read history from {history_file}')


        self.larch = Interpreter(historyfile=history_file,
                                 maxhistory=maxhist)
        self.larch.writer = StdWriter(_larch=self.larch)

        if with_wx and HAS_WXPYTHON:
            symtable = self.larch.symtable
            try:
                from .wxlib import LarchWxApp
                app = LarchWxApp(redirect=False, clearSigInt=False)
            except SystemExit:
                with_wx = False

            symtable.set_symbol('_sys.wx.wxapp', app)
            symtable.set_symbol('_sys.wx.force_wxupdate', False)
            symtable.set_symbol('_sys.wx.parent', None)

            from .wxlib import inputhook
            symtable.set_symbol('_sys.wx.inputhook', inputhook)
            if uname == 'darwin':
                symtable.set_symbol('_sys.wx.ping',   inputhook.ping_darwin)
            else:
                symtable.set_symbol('_sys.wx.ping',   inputhook.ping)

            inputhook.ON_INTERRUPT = self.onCtrlC
            inputhook.WXLARCH_SYM = symtable

        signal.signal(signal.SIGINT, self.onCtrlC)
        self.prompt = self.larch.input.prompt
        writer = self.larch.writer
        self.color_writer = (uname != 'win' and hasattr(writer, 'set_textstyle'))
        if not quiet:
            if banner_msg is None:
                banner_msg = make_banner(show_libraries=['numpy', 'scipy', 'matplotlib', 'h5py',
                                                         'lmfit', 'xraydb', 'wx','wxmplot'])
            if self.color_writer:
                writer.set_textstyle('error')
            writer.write(banner_msg)
            writer.write("\n")
            if self.color_writer:
                writer.set_textstyle('text')

        self.larch_execute = self.default
        self.larch.run_init_scripts()

    def onCtrlC(self, *args, **kws):
        self.larch.symtable.set_symbol('_sys.wx.keyboard_interrupt', True)
        return 0

    def on_exit(self, text=None):
        "exit"
        trim_last = False
        if text is not None:
            trim_last = text.strip() in ('quit', 'exit')
        try:
            self.larch.input.history.save(trim_last=trim_last)
        except PermissionError:
            print("Warning: could not save session history -- permission denied")
        self.larch.symtable._plotter.close_all_displays()
        sys.exit()

    def do_exit(self, text):
        "exit"
        self.on_exit(text=text)

    def do_quit(self, text):
        "quit"
        self.on_exit(text=text)

    def emptyline(self):
        pass

    def onecmd(self, line):
        "single command"
        return self.default(line)

    def do_help(self, arg):
        "help"
        if arg.startswith('(') and arg.endswith(')'):
            arg = arg[1:-1]
        elif arg.startswith("'") and arg.endswith("'"):
            arg = arg[1:-1]
        elif arg.startswith('"') and arg.endswith('"'):
            arg = arg[1:-1]
        self.default(f"help({arg})")

    def do_shell(self, txt):
        "shell command"
        os.system(txt)

    def default(self, line):
        "default handler"
        if line.strip() in ('quit', 'exit', 'quit()', 'exit()', 'EOF'):
            self.on_exit(line)
        ret = self.larch.eval(line, fname='<stdin>', lineno=0)
        if self.larch.error:
            self.larch.input.clear()
            if self.color_writer:
                self.larch.writer.set_textstyle('error')
            self.larch.show_errors()
            if self.color_writer:
                self.larch.writer.set_textstyle('line')
        if ret is not None:
            self.larch.writer.write("%s\n" % repr(ret))

        self.larch.writer.flush()
        self.prompt = self.larch.input.next_prompt



#!/usr/bin/env python
""" Tests of Larch Scripts  """
import unittest
from pathlib import Path
import time
import ast
import numpy as np
import os
from sys import version_info

from utils import TestCase
from larch import Interpreter


base_dir = Path(__file__).parent.parent.resolve()

class TestScripts(TestCase):
    '''tests'''

    def test_basic_interp(self):
        self.runscript('interp.lar', dirname=base_dir / 'examples' / 'basic')
        assert(len(self.session.get_errors()) == 0)
        self.isNear("y0[1]", 0.48578, places=3)
        self.isNear("y1[1]", 0.81310, places=3)
        self.isNear("y2[1]", 0.41532, places=3)

    def test_basic_smooth(self):
        self.runscript('smoothing.lar', dirname=base_dir / 'examples' / 'basic')
        assert(len(self.session.get_errors()) == 0)
        self.isNear("s_loren[5]", 0.087, places=2)
        self.isNear("s_gauss[5]", 3.958e-05, places=2)
        self.isNear("s_voigt[5]", 0.0957, places=2)

    def test_basic_localnames(self):
        self.runscript('local_namespaces.lar', dirname=base_dir / 'examples' / 'basic')
        assert(len(self.session.get_errors()) == 0)
        self.isNear("x", 1000.0, places=4)

    def test_basic_pi(self):
        self.runscript('pi_archimedes.lar', dirname=base_dir / 'examples' / 'basic')
        assert(len(self.session.get_errors()) == 0)
        self.isNear("result", 3.14159265358979267, places=8)

    def test_basic_use_params(self):
        self.runscript('use_params.lar', dirname=base_dir / 'examples' / 'basic')
        assert(len(self.session.get_errors()) == 0)
        self.isNear("a",  0.76863, places=4)

    def test_nested_runfiles(self):
        origdir = Path.cwd().absolute()
        dirname = base_dir / "tests" / "larch_scripts"
        os.chdir(dirname)
        out, err = self.trytext("run('nested_outer.lar')")

        os.chdir(origdir)
        out = out.split('\n')
        assert(len(out) > 4)
        assert('before nested_inner.lar' in out[0])
        assert('in nested_inner.lar' in out[1])
        assert('in nested_deep.lar' in out[2])
        assert('in nested_inner.lar, after nested_deep' in out[3])
        assert('in nested_outer.lar, after nested_inner' in out[4])
        self.isNear("deep_x",  5.0, places=2)

    def test_runfit(self):
        origdir = Path.cwd().absolute()
        dirname = base_dir / "tests" / "larch_scripts"
        os.chdir(dirname)

        out, err = self.trytext("run('fit_constraint.lar')")

        os.chdir(origdir)
        self.isTrue('out.nfev > 30')
        self.isTrue('out.nfev < 70')
        self.isNear('params.amp1.value', 6.05, places=2)
        self.isNear('params.amp2.value', 2.02, places=2)
        self.isNear('params.cen1.value', 3.01, places=2)
        self.isNear('params.cen1.stderr', 0.0073, places=2)
        self.isNear('out.chi_square', 8.4,places=2)


if __name__ == '__main__':  # pragma: no cover
    for suite in (TestScripts,):
        suite = unittest.TestLoader().loadTestsFromTestCase(suite)
        unittest.TextTestRunner(verbosity=13).run(suite)



#!/usr/bin/env python
""" Tests of Larch Scripts  """
import unittest
from pathlib import Path
import time
import ast
import numpy as np
from sys import version_info

from utils import TestCase
from larch import Interpreter


base_dir = Path(__file__).parent.parent.resolve()

class TestScripts(TestCase):
    '''testing of examples/xray'''
    def test1_mu_elam(self):
        self.runscript('get_mu_tables.lar', dirname=base_dir / 'examples' / 'xray')
        assert(len(self.session.get_errors()) == 0)
        self.isTrue("len(mu1) == 401")
        self.isTrue("min(mu1) < 20")

    def test2_mu_elam(self):
        self.session.run("zn_iz = atomic_number('zn')")
        self.session.run("zn_mass = atomic_mass('zn')")
        assert(len(self.session.get_errors()) == 0)
        self.isTrue("zn_iz == 30")
        self.isTrue("zn_mass > 60.")


if __name__ == '__main__':  # pragma: no cover
    for suite in (TestScripts,):
        suite = unittest.TestLoader().loadTestsFromTestCase(suite)
        unittest.TextTestRunner(verbosity=3).run(suite)



# Sphinx extensions for generating PDF/EPUB output
extensions = ['sphinx.ext.imgmath',
              'sphinx.ext.autodoc',
              'sphinx.ext.napoleon',
              'sphinx.ext.extlinks',
              'sphinx.ext.intersphinx',
              'sphinx.ext.ifconfig',
              'sphinxcontrib.bibtex',
              'subfig']



#!/usr/bin/env python
"""
Epics XRF Display App
"""

import sys
import os

import time
import copy
from functools import partial

import wx
import wx.lib.mixins.inspection
import wx.lib.scrolledpanel as scrolled
import wx.dataview as dv

import numpy as np
import matplotlib

from wxmplot import PlotPanel
from wxutils import (SimpleText, EditableListBox, Font, FloatCtrl,
                     pack, Popup, Button, get_icon, Check, MenuItem,
                     Choice, FileOpen, FileSave, fix_filename, HLine,
                     GridPanel, CEN, LEFT, RIGHT)

import larch
from larch.site_config import icondir
from larch.wxlib import PeriodicTablePanel, LarchWxApp
from larch.wxlib.xrfdisplay import (XRFDisplayFrame, XRFCalibrationFrame,
                                    FILE_WILDCARDS)
from larch.utils import get_cwd

ROI_WILDCARD = 'Data files (*.dat)|*.dat|ROI files (*.roi)|*.roi|All files (*.*)|*.*'
try:
    from epics import caget
    from .xrf_detectors import Epics_MultiXMAP, Epics_Xspress3
except:
    pass

class DetectorSelectDialog(wx.Dialog):
    """Connect to an Epics MCA detector
    Can be either XIA xMAP  or Quantum XSPress3
    """
    msg = '''Select XIA xMAP or Quantum XSPress3 MultiElement MCA detector'''
    det_types = ('SXD-7', 'ME-7', 'ME-4', 'other')
    ioc_types = ('Xspress3.1', 'xMAP', 'Xspress3.0')
    def_prefix = '13QX7:'   # SDD1:'
    def_nelem  =  4

    def __init__(self, parent=None, prefix=None, det_type='ME-4',
                 ioc_type='Xspress3', nmca=4,
                 title='Select Epics MCA Detector'):
        if prefix is None:
            prefix = self.def_prefix
        if det_type not in self.det_types:
            det_type = self.det_types[0]

        wx.Dialog.__init__(self, parent, wx.ID_ANY, title=title)

        self.SetBackgroundColour((240, 240, 230))
        self.SetFont(Font(9))
        if parent is not None:
            self.SetFont(parent.GetFont())

        self.ioctype = Choice(self,size=(120, -1), choices=self.ioc_types)
        self.ioctype.SetStringSelection(ioc_type)

        self.dettype = Choice(self,size=(120, -1), choices=self.det_types)
        self.dettype.SetStringSelection(det_type)

        self.prefix = wx.TextCtrl(self, -1, prefix, size=(120, -1))
        self.nelem = FloatCtrl(self, value=nmca, precision=0, minval=1,
                               size=(120, -1))

        btnsizer = wx.StdDialogButtonSizer()

        if wx.Platform != "__WXMSW__":
            btn = wx.ContextHelpButton(self)
            btnsizer.AddButton(btn)

        btn = wx.Button(self, wx.ID_OK)
        btn.SetHelpText("Use this detector")
        btn.SetDefault()
        btnsizer.AddButton(btn)

        btn = wx.Button(self, wx.ID_CANCEL)
        btnsizer.AddButton(btn)
        btnsizer.Realize()

        hline = wx.StaticLine(self, size=(225, 3), style=wx.LI_HORIZONTAL)
        sty = LEFT
        sizer = wx.GridBagSizer(5, 2)
        def txt(label):
            return SimpleText(self, label, size=(120, -1), style=LEFT)

        sizer.Add(txt('  Detector Type'),  (0, 0), (1, 1), sty, 2)
        sizer.Add(txt('  Uses Xspress3?'), (1, 0), (1, 1), sty, 2)
        sizer.Add(txt('  Epics Prefix'),  (2, 0), (1, 1), sty, 2)
        sizer.Add(txt('  # Elements'),    (3, 0), (1, 1), sty, 2)
        sizer.Add(self.dettype,         (0, 1), (1, 1), sty, 2)
        sizer.Add(self.ioctype,         (1, 1), (1, 1), sty, 2)
        sizer.Add(self.prefix,          (2, 1), (1, 1), sty, 2)
        sizer.Add(self.nelem,           (3, 1), (1, 1), sty, 2)

        sizer.Add(hline,                (4, 0), (1, 2), sty, 2)
        sizer.Add(btnsizer,             (5, 0), (1, 2), sty, 2)
        self.SetSizer(sizer)
        sizer.Fit(self)


class EpicsXRFDisplayFrame(XRFDisplayFrame):
    _about = """Epics XRF Spectra Display
  Matt Newville <newville @ cars.uchicago.edu>
  """
    me4_layout = ((0, 0), (1, 0), (1, 1), (0, 1))
    main_title = 'Epics XRF Control'

    def __init__(self, parent=None, _larch=None, prefix=None,
                 det_type='ME-4', ioc_type='Xspress3', nmca=4,
                 size=(725, 580), environ_file=None,
                 title='Epics XRF Display', output_title='XRF', **kws):

        self.det_type = det_type
        self.ioc_type = ioc_type
        self.nmca = nmca
        self.det_fore = 1
        self.det_back = 0
        self.environ = []
        if environ_file is not None:
            self.read_environfile(environ_file)

        self.onConnectEpics(event=None, prefix=prefix)

        self.icon_file = os.path.join(icondir, 'ptable.ico')

        XRFDisplayFrame.__init__(self, parent=parent, _larch=_larch,
                                 title=title, size=size, **kws)

    def read_environfile(self, filename):
        """read environmnet file"""
        if os.path.exists(filename):
            textlines = []
            try:
                with open(filename, 'r') as fh:
                    textlines = fh.readlines()
            except IOError:
                return
            self.environ = []
            for line in textlines:
                line = line[:-1].replace('\t', ' ')
                pvname, desc = line.split(None, 1)
                desc = desc.strip()
                self.environ.append((pvname, desc))

    def onConnectEpics(self, event=None, prefix=None, **kws):
        if prefix is None:
            res  = self.prompt_for_detector(prefix=prefix,
                                            ioc_type=self.ioc_type,
                                            nmca=self.nmca)
            self.prefix, self.det_type, self.ioc_type, self.nmca = res
        else:
            self.prefix = prefix
        self.det_fore = 1
        self.det_back = 0
        self.clear_mcas()
        self.connect_to_detector(prefix=self.prefix, ioc_type=self.ioc_type,
                                 det_type=self.det_type, nmca=self.nmca)

    def onSaveMCAFile(self, event=None, **kws):
        tmp = '''
        # print('SaveMCA File')
        deffile = ''
        if hasattr(self.mca, 'sourcefile'):
            deffile = "%s%s" % (deffile, getattr(self.mca, 'sourcefile'))
        if hasattr(self.mca, 'areaname'):
            deffile = "%s%s" % (deffile, getattr(self.mca, 'areaname'))
        if deffile == '':
            deffile ='test'
        if not deffile.endswith('.mca'):
            deffile = deffile + '.mca'
        '''

        deffile = 'save.mca' # fix_filename(str(deffile))
        outfile = FileSave(self, "Save MCA File",
                           default_file=deffile,
                           wildcard=FILE_WILDCARDS)

        env = []
        if len(self.environ) > 0:
            for pvname, desc in self.environ:
                val  = caget(pvname, as_string=True)
                env.append((pvname, val, desc))

        if outfile is not None:
            self.det.save_mcafile(outfile, environ=env)

    def onSaveColumnFile(self, event=None, **kws):
        print( '  EPICS-XRFDisplay onSaveColumnFile not yet implemented  ')
        pass

    def prompt_for_detector(self, prefix=None, ioc_type='Xspress3',  nmca=4):
        dlg = DetectorSelectDialog(prefix=prefix, ioc_type=ioc_type, nmca=nmca)
        dlg.Raise()
        if dlg.ShowModal() == wx.ID_OK:
            dpref = dlg.prefix.GetValue()
            atype = dlg.ioctype.GetStringSelection()
            dtype = dlg.dettype.GetStringSelection()
            nmca = dlg.nelem.GetValue()
            dlg.Destroy()
        return dpref, dtype, atype, nmca

    def connect_to_detector(self, prefix=None, ioc_type='Xspress3',
                            det_type=None, nmca=4):
        self.det = None
        ioc_type = ioc_type.lower()
        if ioc_type.startswith('xspress3'):
            version = 2
            if 'old' in ioc_type:
                version = 1
            self.det = Epics_Xspress3(prefix=prefix, nmca=nmca, version=version)
            self.det.connect()
            time.sleep(0.5)
            self.det.get_mca(mca=1)
            self.needs_newplot=True
        else:
            self.det = Epics_MultiXMAP(prefix=prefix, nmca=nmca)
        time.sleep(0.05)

    def show_mca(self, init=False):
        self.needs_newplot = False
        if self.mca is None or self.needs_newplot:
            self.mca = self.det.get_mca(mca=self.det_fore)

        self.plotmca(self.mca, set_title=False, init=init)
        title = "Foreground: MCA{:d}".format(self.det_fore)
        if self.det_back  > 0:
            if self.mca2 is None:
                self.mca2 = self.det.get_mca(mca=self.det_back)

            c2 = self.det.get_array(mca=self.det_back)
            e2 = self.det.get_energy(mca=self.det_back)
            title = "{:s}  Background: MCA{:d}".format(title, self.det_back)
            try:
                self.oplot(e2, c2)
            except ValueError:
                pass

        roiname = self.get_roiname()

        if roiname in self.wids['roilist'].GetStrings():
            i = self.wids['roilist'].GetStrings().index(roiname)
            self.wids['roilist'].EnsureVisible(i)
            self.onROI(label=roiname)
        deadtime = self.det.get_deadtime(mca=self.det_fore)
        if deadtime is not None:
            self.wids['deadtime'].SetLabel("%.1f" % deadtime)
        self.SetTitle("%s: %s" % (self.main_title, title))
        self.needs_newplot = False

    def onSaveROIs(self, event=None, **kws):
        dlg = wx.FileDialog(self, message="Save ROI File",
                            defaultDir=get_cwd(),
                            wildcard=ROI_WILDCARD,
                            style = wx.FD_SAVE|wx.FD_CHANGE_DIR)

        if dlg.ShowModal() == wx.ID_OK:
            roifile = dlg.GetPath()

        self.det.save_rois(roifile)

    def onRestoreROIs(self, event=None, **kws):
        dlg = wx.FileDialog(self, message="Read ROI File",
                            defaultDir=get_cwd(),
                            wildcard=ROI_WILDCARD,
                            style = wx.FD_OPEN|wx.FD_CHANGE_DIR)

        if dlg.ShowModal() == wx.ID_OK:
            roifile = dlg.GetPath()
            self.det.restore_rois(roifile)
            self.set_roilist(mca=self.mca)
            self.show_mca()
            self.onSelectDet(event=None, index=0)

    def createCustomMenus(self):
        menu = wx.Menu()
        MenuItem(self, menu, "Connect to Detector\tCtrl+D",
                 "Connect to MCA or XSPress3 Detector",
                 self.onConnectEpics)
        menu.AppendSeparator()
        self._menus.insert(1, (menu, 'Detector'))

    def createMainPanel(self):
        epicspanel = self.createEpicsPanel()
        ctrlpanel  = self.createControlPanel()
        plotpanel  = self.panel = self.createPlotPanel()
        self.panel.SetName('plotpanel')
        tx, ty = self.wids['ptable'].GetBestSize()
        cx, cy = ctrlpanel.GetBestSize()
        px, py = plotpanel.GetBestSize()

        self.SetSize((950, 625))
        self.SetMinSize((450, 350))

        style = wx.ALIGN_LEFT|wx.EXPAND|wx.ALL

        bsizer = wx.BoxSizer(wx.HORIZONTAL)
        bsizer.Add(ctrlpanel, 0, style, 1)
        bsizer.Add(plotpanel, 1, style, 1)
        hline = wx.StaticLine(self, size=(425, 2), style=wx.LI_HORIZONTAL|style)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(epicspanel, 0, style, 1)
        sizer.Add(hline,      0, style, 1)
        sizer.Add(bsizer,     1, style, 1)
        pack(self, sizer)

        try:
            self.SetIcon(wx.Icon(self.icon_file, wx.BITMAP_TYPE_ICO))
        except:
            pass
        self.set_roilist(mca=None)

    def create_detbuttons(self, pane):
        btnpanel = wx.Panel(pane, name='buttons')
        btnsizer = wx.GridBagSizer(1, 1)
        btns = {}
        sx = 36
        sy = int(sx/2)
        for i in range(1, self.nmca+1):
            b = Button(btnpanel, f'{i}', size=(sx, sx),
                       action=partial(self.onSelectDet, index=i))
            b.SetFont(Font(9))
            self.wids['det%i' % i] = b
            btns[i] = b
        dtype = self.det_type.lower().replace('-', '').replace(' ', '').replace('_', '')
        if dtype.startswith('sxd7') and self.nmca == 7:
            btnsizer.Add((sx, sy), (0, 0), (1, 2), wx.ALIGN_LEFT, 1)
            btnsizer.Add(btns[6],  (1, 0), (2, 2), wx.ALIGN_LEFT, 1)
            btnsizer.Add(btns[7],  (3, 0), (2, 2), wx.ALIGN_LEFT, 1)
            btnsizer.Add((sx, sy), (5, 0), (1, 2), wx.ALIGN_LEFT, 1)
            btnsizer.Add(btns[5],  (0, 2), (2, 2), wx.ALIGN_LEFT, 1)
            btnsizer.Add(btns[4],  (2, 2), (2, 2), wx.ALIGN_LEFT, 1)
            btnsizer.Add(btns[1],  (4, 2), (2, 2), wx.ALIGN_LEFT, 1)
            btnsizer.Add((sx, sy), (0, 4), (1, 2), wx.ALIGN_LEFT, 1)
            btnsizer.Add(btns[3],  (1, 4), (2, 2), wx.ALIGN_LEFT, 1)
            btnsizer.Add(btns[2],  (3, 4), (2, 2), wx.ALIGN_LEFT, 1)
            btnsizer.Add((sx, sy), (5, 4), (1, 2), wx.ALIGN_LEFT, 1)
        elif dtype.startswith('me7') and self.nmca == 7:
            btnsizer.Add((sx, sy), (0, 0), (1, 2), wx.ALIGN_LEFT, 1)
            btnsizer.Add(btns[7],  (1, 0), (2, 2), wx.ALIGN_LEFT, 1)
            btnsizer.Add(btns[6],  (3, 0), (2, 2), wx.ALIGN_LEFT, 1)
            btnsizer.Add((sx, sy), (5, 0), (1, 2), wx.ALIGN_LEFT, 1)
            btnsizer.Add(btns[2],  (0, 2), (2, 2), wx.ALIGN_LEFT, 1)
            btnsizer.Add(btns[1],  (2, 2), (2, 2), wx.ALIGN_LEFT, 1)
            btnsizer.Add(btns[5],  (4, 2), (2, 2), wx.ALIGN_LEFT, 1)
            btnsizer.Add((sx, sy), (0, 4), (1, 2), wx.ALIGN_LEFT, 1)
            btnsizer.Add(btns[3],  (1, 4), (2, 2), wx.ALIGN_LEFT, 1)
            btnsizer.Add(btns[4],  (3, 4), (2, 2), wx.ALIGN_LEFT, 1)
            btnsizer.Add((sx, sy), (5, 4), (1, 2), wx.ALIGN_LEFT, 1)
        elif dtype.startswith('me4') and self.nmca == 4:
            btnsizer.Add(btns[1],  (0, 0), (1, 1), wx.ALIGN_LEFT, 1)
            btnsizer.Add(btns[2],  (1, 0), (1, 1), wx.ALIGN_LEFT, 1)
            btnsizer.Add(btns[3],  (1, 1), (1, 1), wx.ALIGN_LEFT, 1)
            btnsizer.Add(btns[4],  (0, 1), (1, 1), wx.ALIGN_LEFT, 1)
        else:
            NPERROW = 4
            icol, irow = 0, 0
            for nmca  in range(1, self.nmca+1):
                btnsizer.Add(btns[nmca],  (irow, icol), (1, 1), wx.ALIGN_LEFT, 1)
                icol += 1
                if icol > NPERROW-1:
                    icol = 0
                    irow += 1

        pack(btnpanel, btnsizer)
        return btnpanel

    def createEpicsPanel(self):
        pane = wx.Panel(self, name='epics panel')
        style  = wx.ALIGN_LEFT
        rstyle = wx.ALIGN_RIGHT

        det_btnpanel = self.create_detbuttons(pane)

        bkg_choices = ['None'] + ["%d" % (i+1) for i in range(self.nmca)]

        self.wids['det_status'] = SimpleText(pane, ' ', size=(120, -1), style=style)
        self.wids['deadtime']   = SimpleText(pane, ' ', size=(120, -1), style=style)

        self.wids['bkg_det'] = Choice(pane, size=(100, -1), choices=bkg_choices,
                                      action=self.onSelectDet)

        self.wids['dwelltime'] = FloatCtrl(pane, value=0.0, precision=1, minval=0,
                                           size=(80, -1), act_on_losefocus=True,
                                           action=self.onSetDwelltime)
        self.wids['elapsed'] = SimpleText(pane, ' ', size=(80, -1),  style=style)

        self.wids['mca_sum'] = Choice(pane, size=(100, -1),
                                      choices=['Single', 'Accumulate'],
                                      action=self.onMcaSumChoice,
                                      default=1 )

        b1 =  Button(pane, 'Start',      size=(90, -1), action=self.onStart)
        b2 =  Button(pane, 'Stop',       size=(90, -1), action=self.onStop)
        b3 =  Button(pane, 'Erase',      size=(90, -1), action=self.onErase)
        b4 =  Button(pane, 'Continuous', size=(90, -1), action=partial(self.onStart,
                                                                       dtime=0.0))

        sum_lab = SimpleText(pane, 'Accumulate Mode:',   size=(150, -1))
        bkg_lab = SimpleText(pane, 'Background MCA:',   size=(150, -1))
        pre_lab = SimpleText(pane, 'Dwell Time (s):',   size=(125, -1))
        ela_lab = SimpleText(pane, 'Elapsed Time (s):', size=(125, -1))
        sta_lab = SimpleText(pane, 'Status :',          size=(100, -1))
        dea_lab = SimpleText(pane, '% Deadtime:',       size=(100, -1))

        psizer = wx.GridBagSizer(5, 5)
        psizer.Add(SimpleText(pane, ' MCAs: '),  (0, 0), (1, 1), style, 1)
        psizer.Add(det_btnpanel,           (0, 1), (2, 1), style, 1)
        psizer.Add(bkg_lab,                (0, 2), (1, 1), style, 1)
        psizer.Add(self.wids['bkg_det'],   (0, 3), (1, 1), style, 1)
        psizer.Add(sum_lab,                (1, 2), (1, 1), style, 1)
        psizer.Add(self.wids['mca_sum'],   (1, 3), (1, 1), style, 1)
        psizer.Add(pre_lab,                (0, 4), (1, 1),  style, 1)
        psizer.Add(ela_lab,                (1, 4), (1, 1),  style, 1)
        psizer.Add(self.wids['dwelltime'], (0, 5), (1, 1),  style, 1)
        psizer.Add(self.wids['elapsed'],   (1, 5), (1, 1),  style, 1)

        psizer.Add(b1, (0, 6), (1, 1), style, 1)
        psizer.Add(b4, (0, 7), (1, 1), style, 1)
        psizer.Add(b2, (1, 6), (1, 1), style, 1)
        psizer.Add(b3, (1, 7), (1, 1), style, 1)

        psizer.Add(sta_lab,                  (0, 8), (1, 1), style, 1)
        psizer.Add(self.wids['det_status'],  (0, 9), (1, 1), style, 1)
        psizer.Add(dea_lab,                  (1, 8), (1, 1), style, 1)
        psizer.Add(self.wids['deadtime'],    (1, 9), (1, 1), style, 1)
        pack(pane, psizer)
        # pane.SetMinSize((500, 53))
        self.det.connect_displays(status=self.wids['det_status'],
                                  elapsed=self.wids['elapsed'])

        wx.CallAfter(self.onSelectDet, index=1, init=True)
        self.timer_counter = 0
        self.mca_timer = wx.Timer(self)
        self.Bind(wx.EVT_TIMER, self.UpdateData, self.mca_timer)
        self.mca_timer.Start(250)
        return pane

    def UpdateData(self, event=None, force=False):
        self.timer_counter += 1
        if self.mca is None or self.needs_newplot:
            self.show_mca()
        # self.elapsed_real = self.det.elapsed_real
        self.mca.real_time = self.det.elapsed_real
        # print("Update Data  ", force, self.det.needs_refresh)

        if force or self.det.needs_refresh:
            self.det.needs_refresh = False
            if self.det_back > 0:
                if self.mca2 is None:
                    self.mca2 = self.det.get_mca(mca=self.det_back)

                counts = self.det.get_array(mca=self.det_back)
                energy = self.det.get_energy(mca=self.det_back)
                try:
                    self.update_mca(counts, energy=energy, is_mca2=True, draw=False)
                except ValueError:
                    pass

            if self.mca is None:
                self.mca = self.det.get_mca(mca=self.det_fore)

            dtime = self.det.get_deadtime(mca=self.det_fore)
            if dtime is not None:
                self.wids['deadtime'].SetLabel("%.1f" % dtime)

            counts = self.det.get_array(mca=self.det_fore)*1.0
            energy = self.det.get_energy(mca=self.det_fore)
            if max(counts) < 1.0:
                counts    = 1e-4*np.ones(len(counts))
                counts[0] = 2.0
            self.update_mca(counts, energy=energy)

    def ShowROIStatus(self, left, right, name='', panel=0):
        if left > right:
            return
        sum = self.ydata[left:right].sum()

        try:
            ftime, nframes = self.det.get_frametime()
        except:
            ftime   = self.det.frametime
            nframes = self.det.nframes
        self.det.elapsed_real = nframes * ftime

        mca_counts = self.det.mcas[self.det_fore-1].get('VAL')
        sum =  mca_counts[left:right].sum()
        # print("ROI STATUS ", name, ftime, nframes, sum, cps, mca_counts.sum(),  mca_counts)
        if name in (None, ''):
            name = 'Selected'
        else:
            for roi in self.det.mcas[self.det_fore-1].rois:
                if name.lower() == roi.name.lower():
                    try:
                        sum = roi.sum
                    except:
                        pass
        cps = sum/ftime
        if cps < 0: cps = 0
        # print("ROI STATUS ", name, _counts, cps)
        fmt = " {:s}: Cts={:10,.0f} :{:10,.1f} Hz"
        self.write_message(fmt.format(name, sum, cps), panel=panel)

    def onSelectDet(self, event=None, index=0, init=False, **kws):
        if index > 0:
            self.det_fore = index
        self.det_back = self.wids['bkg_det'].GetSelection()
        if self.det_fore  == self.det_back:
            self.det_back = 0

        for i in range(1, self.nmca+1):
            dname = 'det%i' % i
            bcol = (210, 210, 210)
            fcol = (0, 0, 0)
            if i == self.det_fore:
                fcol = (200,  20,  20)
                bcol = (250, 250, 250)
            self.wids[dname].SetBackgroundColour(bcol)
            self.wids[dname].SetForegroundColour(fcol)
        self.clear_mcas()
        self.show_mca(init=init)
        self.Refresh()

    def swap_mcas(self, event=None):
        if self.mca2 is None:
            return
        self.mca, self.mca2 = self.mca2, self.mca
        fore, back = self.det_fore, self.det_back
        self.wids['bkg_det'].SetSelection(fore)
        self.onSelectDet(index=back)

    def onMcaSumChoice(self, event=None):
        wid = self.wids['mca_sum']
        self.det.set_usesum('accum' in wid.GetStringSelection().lower())

    def onSetDwelltime(self, event=None, **kws):
        if 'dwelltime' in self.wids:
            self.det.set_dwelltime(dtime=self.wids['dwelltime'].GetValue())

    def clear_mcas(self):
        self.mca = self.mca2 = None
        self.x2data = self.y2data = None
        self.needs_newplot = True

    def onStart(self, event=None, dtime=None, **kws):
        if dtime is not None:
            self.wids['dwelltime'].SetValue("%.1f" % dtime)
            self.det.set_dwelltime(dtime=dtime)
        else:
            self.det.set_dwelltime(dtime=self.wids['dwelltime'].GetValue())
        self.det.start()

    def onStop(self, event=None, **kws):
        self.det.stop()
        self.det.needs_refresh = True
        time.sleep(0.05)
        self.UpdateData(event=None, force=True)

    def onErase(self, event=None, **kws):
        self.needs_newplot = True
        self.det.erase()

    def onDelROI(self, event=None):
        roiname = self.get_roiname()
        errmsg = None
        t0 = time.time()
        if self.roilist_sel is None:
            errmsg = 'No ROI selected to delete.'
        if errmsg is not None:
            return Popup(self, errmsg, 'Cannot Delete ROI')

        self.det.del_roi(roiname)
        XRFDisplayFrame.onDelROI(self)


    def onNewROI(self, event=None):
        roiname = self.get_roiname()
        errmsg = None
        if self.xmarker_left is None or self.xmarker_right is None:
            errmsg = 'Must select right and left markers to define ROI'
        elif roiname in self.wids['roilist'].GetStrings():
            errmsg = '%s is already in ROI list - use a unique name.' % roiname
        if errmsg is not None:
            return Popup(self, errmsg, 'Cannot Define ROI')

        confirmed = XRFDisplayFrame.onNewROI(self)
        if confirmed:
            self.det.add_roi(roiname, lo=self.xmarker_left,
                             hi=self.xmarker_right)

    def onRenameROI(self, event=None):
        roiname = self.get_roiname()
        errmsg = None
        if roiname in self.wids['roilist'].GetStrings():
            errmsg = '%s is already in ROI list - use a unique name.' % roiname
        elif self.roilist_sel is None:
            errmsg = 'No ROI selected to rename.'
        if errmsg is not None:
            return Popup(self, errmsg, 'Cannot Rename ROI')

        if self.roilist_sel < len(self.det.mcas[0].rois):
            self.det.rename_roi(self.roilist_sel, roiname)
            names = self.wids['roilist'].GetStrings()
            names[self.roilist_sel] = roiname
            self.wids['roilist'].Clear()
            for sname in names:
                self.wids['roilist'].Append(sname)
            self.wids['roilist'].SetSelection(self.roilist_sel)

    def onCalibrateEnergy(self, event=None, **kws):
        try:
            self.win_calib.Raise()
        except:
            self.win_calib = XRFCalibrationFrame(self, mca=self.mca,
                                              larch=self.larch,
                                              callback=self.onSetCalib)

    def onSetCalib(self, offset, slope, mca=None):
        print('XRFControl Set Energy Calibratione' , offset, slope, mca)

    def onClose(self, event=None):
        self.onStop()
        XRFDisplayFrame.onClose(self)

    def onExit(self, event=None):
        self.onStop()
        XRFDisplayFrame.onExit(self)

class EpicsXRFApp(LarchWxApp):
    def __init__(self, _larch=None, prefix=None,
                 det_type='ME-4', ioc_type='Xspress3', nmca=4,
                 size=(725, 580), environ_file=None,
                 title='Epics XRF Display', output_title='XRF', **kws):
        self.prefix = prefix
        self.det_type = det_type
        self.ioc_type = ioc_type
        self.nmca = nmca
        self.size = size
        self.environ_file = environ_file
        self.title = title
        self.output_title = output_title
        LarchWxApp.__init__(self, _larch=_larch, **kws)

    def createApp(self):
        frame = EpicsXRFDisplayFrame(prefix=self.prefix,
                                     det_type=self.det_type,
                                     ioc_type=self.ioc_type,
                                     nmca=self.nmca, size=self.size,
                                     environ_file=self.environ_file,
                                     title=self.title,
                                     output_title=self.output_title,
                                     _larch=self._larch)
        frame.Show()
        self.SetTopWindow(frame)
        return True

if __name__ == "__main__":
    EpicsXRFApp().MainLoop()



#!/usr/bin/env python
"""Version information"""

__release_version__ = '0.9.81'
__date__    = '2024-September-20'
__authors__ = "M. Newville, M. Rovezzi, M. Koker, B. Ravel, and others"

from ._version import __version__, __version_tuple__

import sys
from collections import namedtuple
from packaging.version import parse as ver_parse
import importlib
import urllib3
import requests
import numpy
import scipy
import matplotlib
import lmfit

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

# libraries whose versions might be interesting to know
LIBS_VERSIONS = ('numpy', 'scipy', 'matplotlib', 'h5py', 'sklearn', 'skimage',
                 'sqlalchemy', 'fabio', 'pyFAI', 'PIL', 'imageio', 'silx',
                 'tomopy', 'pymatgen.core', 'numdifftools', 'xraydb', 'lmfit',
                 'asteval', 'wx', 'wxmplot')

def version_data(with_libraries=False):
    "get version data"
    vinf  = sys.version_info
    pyvers = f'{vinf.major:d}.{vinf.minor:d}.{vinf.micro:d}'

    buildinfo = sys.version
    if '\n' in buildinfo:
        buildinfo = buildinfo.split('\n', maxsplit=1)[0].strip()
    if buildinfo.startswith(pyvers):
        buildinfo = buildinfo.replace(pyvers, '').strip()

    builder = buildinfo[:40]
    if '|' in buildinfo:
        sects = buildinfo.split('|')
        if len(sects) > 1:
            builder = sects[1].strip()

    vdat = {'release version': __release_version__,
            'release date':  __date__,
            'development version':  __version__,
            'authors': __authors__,
            'python version': pyvers,
            'python builder': builder,
            'python sysversion': sys.version,
            }

    if with_libraries:
        for modname in LIBS_VERSIONS:
            vers = "not installed"
            if modname not in sys.modules:
                try:
                    importlib.import_module(modname)
                except:
                    pass
            if modname in sys.modules:
                mod = sys.modules[modname]
                vers = getattr(mod, '__version__', None)
                if vers is None:
                    vers = getattr(mod, 'version',
                                   'unknown version')
            vdat[modname] = vers
    return vdat

def make_banner(with_libraries=False, show_libraries=None):
    "return startup banner"
    if show_libraries is None:
        show_libraries = LIBS_VERSIONS if with_libraries else []

    vdat = version_data(with_libraries=True)
    lines = [f"Larch {vdat['release version']}, released {vdat['release date']}"]
    if vdat['development version'] != vdat['release version']:
        lines.append(f"development version: {vdat['development version']}")
    lines.append(f"Python {vdat['python version']}, {vdat['python builder']}")
    libline = []
    for key, val in vdat.items():
        if key in show_libraries:
            libline.append(f"{key:s}: {val}")
            if len(libline) > 3:
                lines.append(', '.join(libline))
                libline = []
    if len(libline) > 0:
        lines.append(', '.join(libline))

    if len(show_libraries) < 10:
        lines.append('use `print(show_version())` for detailed versions')

    linelen = max([len(line) for line in lines])
    border = '='*min(99, max(linelen, 25))
    lines.insert(0, border)
    lines.append(border)
    return '\n'.join(lines)

def show_version():
    vdat = version_data(with_libraries=True)
    out = []
    for key, val in vdat.items():
        out.append(f"{key:20s}:  {val}")
    return '\n'.join(out)


########
## for comparing current with remote version
########
VERSION_URL='https://raw.githubusercontent.com/xraypy/xraylarch/gh-pages/version.txt'

VersionStatus = namedtuple('VersionStatus', ('update_available',
                                             'local_version',
                                             'remote_version',
                                             'message'))

UPDATE_MESSAGE = """#=== Update Available ===
Larch version {remote_version:s} is available. Your version is currently {local_version:s}.
To update the latest version run
   larch -u
from a Command Window or Terminal.
#========================"""

LATEST_MESSAGE = """Larch version {local_version:s} is up to date."""

def check_larchversion():
    "check version, return VersionStatus tuple"
    local_version = __release_version__

    try:
        req = requests.get(VERSION_URL, verify=False, timeout=3.10)
    except:
        return VersionStatus(False, local_version, 'unknown', 'offline')
    remote_version = '0.9.001'
    if req.status_code == 200:
        try:
            for line in req.text.split('\n'):
                line = line.strip()
                if not line.startswith('#'):
                    remote_version = line
                    break
        except:
            pass
    update_available = ver_parse(remote_version) > ver_parse(local_version)
    message = UPDATE_MESSAGE if update_available else LATEST_MESSAGE
    message = message.format(remote_version=remote_version,
                             local_version=local_version)
    return VersionStatus(update_available,local_version, remote_version, message)



#!/usr/bin/env python

"""
Standard plots for Larch
------------------------

This module is independent of the Wx GUI. It uses matplotlib or Plotly as
plotting libraries. It is intended as plotter module for standalone scripts and
Jupyter notebooks.
"""

from . import plotly_xafsplots



## examples/pca/pca_aucyano.py
# note that this is similar to examples/fitting/doc_example3

from larch.io import read_athena
from larch.xafs import pre_edge
from larch.math import pca_train, pca_fit

from larch.wxlib.xafsplots import plot_pca_components, plot_pca_weights, plot_pca_fit

prj = read_athena('cyanobacteria.prj', do_fft=False, do_bkg=False)


standards = (prj.Au_foil, prj.Au3_Cl_aq, prj.Au_hydroxide, prj.Au_thiocyanide,
             prj.Au_sulphide, prj.Au_thiosulphate_aq)

unknowns = (prj.d_0_12, prj.d_2_42, prj.d_4_73, prj.d_7_03,
           prj.d_9_33, prj.d_20, prj.d_33, prj.d_720)

# make sure pre_edge() is run with the same params for all groups
for grp in standards + unknowns:
    pre_edge(grp, pre1=-150, pre2=-30, nnorm=1, norm1=150, norm2=850)
#endfor

# train model with standards
au_pcamodel = pca_train(standards, arrayname='norm', xmin=11870, xmax=12030)


# plot components and weights
plot_pca_components(au_pcamodel, min_weight=0.005)
plot_pca_weights(au_pcamodel, win=2, min_weight=0.005, ylog_scale=True)

# print out weights
total = 0
print(" Comp #  |  Weight   |  Cumulative Total")
for i, weight in enumerate(au_pcamodel.variances):
    total = total + weight
    print("  %3i    | %8.5f  | %8.5f " % (i+1, weight, total))
#endfor

# fit unknown data to model
pca_fit(prj.d_720, au_pcamodel, ncomps=4)

# plot
plot_pca_fit(prj.d_720, win=3)


## end of examples/pca/pca_aucyano.py



from contextlib import contextmanager
from typing import Iterator, Optional
import h5py

H5PY_VERSION = h5py.version.version_tuple[:3]
H5PY_HAS_LOCKING = H5PY_VERSION >= (3, 5)


@contextmanager
def open(filename) -> Iterator[h5py.File]:
    kw = {"mode": "r"}
    if H5PY_HAS_LOCKING:
        kw["locking"] = False
    with h5py.File(filename, **kw) as f:
        yield f


def nexus_creator(filename: str) -> str:
    with open(filename) as nxroot:
        return nxroot.attrs.get("creator", "")


def nexus_instrument(filename: str) -> str:
    with open(filename) as nxroot:
        entry = find_nexus_class(nxroot, "NXentry")
        if entry is None:
            return ""

        instrument = find_nexus_class(entry, "NXinstrument")
        if instrument is None:
            return ""

        if "name" in instrument:
            return asstr(instrument["name"][()])
    return ""


def nexus_source(filename: str) -> str:
    with open(filename) as nxroot:
        entry = find_nexus_class(nxroot, "NXentry")
        if entry is None:
            return ""

        source = find_nexus_class(entry, "NXsource")
        if source is None:
            instrument = find_nexus_class(entry, "NXinstrument")
            if instrument is None:
                return ""
            source = find_nexus_class(instrument, "NXsource")
            if source is None:
                return ""

        if "name" in source:
            return asstr(source["name"][()])
    return ""


def asstr(s):
    if isinstance(s, bytes):
        return s.decode()
    return s


def find_nexus_class(parent: h5py.Group, nxclass: str) -> Optional[h5py.Group]:
    for name in parent:
        try:
            child = parent[name]
        except KeyError:
            continue  # broken line
        if asstr(child.attrs.get("NX_class", "")) != nxclass:
            continue
        return child



"""
Fixes equation references from Sphinx math domain
from Equation (1) to Equation 1, which is what they
should be. Must be before sphinx.ext.math* in
extensions list.
"""

from docutils import nodes
import sphinx.ext.mathbase
from sphinx.ext.mathbase import displaymath, eqref

def number_equations(app, doctree, docname):
    num = 0
    numbers = {}
    for node in doctree.traverse(displaymath):
        if node['label'] is not None:
            num += 1
            node['number'] = num
            numbers[node['label']] = num
        else:
            node['number'] = None
    for node in doctree.traverse(eqref):
        if node['target'] not in numbers:
            continue
        num = '%d' % numbers[node['target']]
        node[0] = nodes.Text(num, num)

sphinx.ext.mathbase.number_equations = number_equations

def setup(app):
    pass



"""
Changes section references to be the section number
instead of the title of the section.
"""

from docutils import nodes
import sphinx.domains.std

class CustomStandardDomain(sphinx.domains.std.StandardDomain):

    def __init__(self, env):
        env.settings['footnote_references'] = 'superscript'
        sphinx.domains.std.StandardDomain.__init__(self, env)

    def resolve_xref(self, env, fromdocname, builder,
                     typ, target, node, contnode):
        res = super(CustomStandardDomain, self).resolve_xref(env, fromdocname, builder,
                                                            typ, target, node, contnode)

        if res is None:
            return res

        if typ == 'ref' and not node['refexplicit']:
            docname, labelid, sectname = self.data['labels'].get(target, ('','',''))
            res['refdocname'] = docname

        return res

def doctree_resolved(app, doctree, docname):
    secnums = app.builder.env.toc_secnumbers
    for node in doctree.traverse(nodes.reference):
        if 'refdocname' in node:
            refdocname = node['refdocname']
            if refdocname in secnums:
                secnum = secnums[refdocname]
                toclist = app.builder.env.tocs[refdocname]
                for child in node.traverse():
                    if isinstance(child, nodes.Text):
                        text = child.astext()
                        anchorname = None
                        for refnode in toclist.traverse(nodes.reference):
                            if refnode.astext() == text:
                                anchorname = refnode['anchorname']
                        if anchorname is None:
                            continue
                        linktext = '.'.join(map(str, secnum[anchorname]))
                        child.parent.replace(child, nodes.Text(linktext))

def setup(app):
    app.override_domain(CustomStandardDomain)
    app.connect('doctree-resolved', doctree_resolved)
